import asyncio
from datetime import datetime

import pytest

from bson import ObjectId

# Check if the required dependancies are met to run this driver's tests
try:
    from motor.motor_asyncio import AsyncIOMotorClient
    from motor import version_tuple as MOTOR_VERSION
except ImportError as e:
    dep_error = str(e)
else:
    dep_error = None

from ..common import BaseDBTest, TEST_DB

from umongo import Document, EmbeddedDocument, fields, exceptions, Reference

if not dep_error:  # Make sure the module is valid by importing it
    from umongo.frameworks import motor_asyncio as framework  # noqa
    from pymongo.results import InsertOneResult, UpdateResult, DeleteResult


def _stripped(indexes):
    # With pymongo==2.8 a `ns` field is returned with Mongodb>=3 but
    # not with MongoDB<2, thus we have to clean this before doing comparing
    # Version may differ between database versions and configurations so it
    # shall not be checked
    return {
        k: {sk: sv for sk, sv in v.items() if sk not in ('ns', 'v')}
        for k, v in indexes.items()
    }


# Helper to sort indexes by name in order to have deterministic comparison
def name_sorted(indexes):
    return sorted(indexes, key=lambda x: x['name'])


def make_db():
    return AsyncIOMotorClient()[TEST_DB]


@pytest.fixture
def db():
    return make_db()


@pytest.fixture
def loop():
    return asyncio.get_event_loop()


@pytest.mark.skipif(dep_error is not None, reason=dep_error)
class TestMotorAsyncio(BaseDBTest):

    def test_create(self, loop, classroom_model):
        Student = classroom_model.Student

        async def do_test():
            john = Student(name='John Doe', birthday=datetime(1995, 12, 12))
            ret = await john.commit()
            assert isinstance(ret, InsertOneResult)
            assert john.to_mongo() == {
                '_id': john.id,
                'name': 'John Doe',
                'birthday': datetime(1995, 12, 12)
            }

            john2 = await Student.find_one(john.id)
            assert john2._data == john._data
            # Double commit should do nothing
            ret = await john.commit()
            assert ret is None

        loop.run_until_complete(do_test())

    def test_update(self, loop, classroom_model):
        Student = classroom_model.Student

        async def do_test():
            john = Student(name='John Doe', birthday=datetime(1995, 12, 12))
            await john.commit()
            john.name = 'William Doe'
            assert john.to_mongo(update=True) == {'$set': {'name': 'William Doe'}}
            ret = await john.commit()
            assert isinstance(ret, UpdateResult)
            assert john.to_mongo(update=True) is None
            john2 = await Student.find_one(john.id)
            assert john2._data == john._data
            # Update without changing anything
            john.name = john.name
            await john.commit()
            # Test conditional commit
            john.name = 'Zorro Doe'
            with pytest.raises(exceptions.UpdateError):
                await john.commit(conditions={'name': 'Bad Name'})
            await john.commit(conditions={'name': 'William Doe'})
            await john.reload()
            assert john.name == 'Zorro Doe'
            # Cannot use conditions when creating document
            with pytest.raises(RuntimeError):
                await Student(name='Joe').commit(conditions={'name': 'dummy'})

        loop.run_until_complete(do_test())

    def test_remove(self, loop, classroom_model):
        Student = classroom_model.Student

        async def do_test():
            await Student.collection.drop()
            john = Student(name='John Doe', birthday=datetime(1995, 12, 12))
            with pytest.raises(exceptions.NotCreatedError):
                await john.remove()
            await john.commit()
            assert (await Student.count_documents()) == 1
            ret = await john.remove()
            assert isinstance(ret, DeleteResult)
            assert not john.is_created
            assert (await Student.count_documents()) == 0
            with pytest.raises(exceptions.NotCreatedError):
                await john.remove()
            # Can re-commit the document in database
            await john.commit()
            assert john.is_created
            assert (await Student.count_documents()) == 1
            # Test conditional delete
            with pytest.raises(exceptions.DeleteError):
                await john.remove(conditions={'name': 'Bad Name'})
            await john.remove(conditions={'name': 'John Doe'})
            # Finally try to remove a doc no longer in database
            await john.commit()
            await (await Student.find_one(john.id)).remove()
            with pytest.raises(exceptions.DeleteError):
                await john.remove()

        loop.run_until_complete(do_test())

    def test_reload(self, loop, classroom_model):
        Student = classroom_model.Student

        async def do_test():
            await Student(name='Other dude').commit()
            john = Student(name='John Doe', birthday=datetime(1995, 12, 12))
            with pytest.raises(exceptions.NotCreatedError):
                await john.reload()
            await john.commit()
            john2 = await Student.find_one(john.id)
            john2.name = 'William Doe'
            await john2.commit()
            await john.reload()
            assert john.name == 'William Doe'

        loop.run_until_complete(do_test())

    def test_cursor(self, loop, classroom_model):
        Student = classroom_model.Student

        async def do_test():
            await Student.collection.drop()

            for i in range(10):
                await Student(name='student-%s' % i).commit()
            cursor = Student.find(limit=5, skip=6)
            if MOTOR_VERSION < (2, 0):
                # this test doesn't make as much sense in motor>2.0, since the cursor no
                # longer supports a count() method
                count = await cursor.count()
                assert count == 10
                count_with_limit_and_skip = await cursor.count(with_limit_and_skip=True)
                assert count_with_limit_and_skip == 4
            else:
                assert (await Student.count_documents()) == 10
                assert (await Student.count_documents(limit=5, skip=6)) == 4

            # to_list with callback should fail
            error = NotImplementedError if MOTOR_VERSION < (2, 0, 0) else TypeError
            with pytest.raises(error):
                await cursor.to_list(length=100, callback=lambda r, e: r if r else e)

            # Make sure returned documents are wrapped
            names = []
            for elem in (await cursor.to_list(length=100)):
                assert isinstance(elem, Student)
                names.append(elem.name)
            assert sorted(names) == ['student-%s' % i for i in range(6, 10)]

            # Try with fetch_next as well
            names = []
            cursor.rewind()
            while (await cursor.fetch_next):
                elem = cursor.next_object()
                assert isinstance(elem, Student)
                names.append(elem.name)
            assert sorted(names) == ['student-%s' % i for i in range(6, 10)]

            # Try with each as well
            names = []
            cursor.rewind()
            future = asyncio.Future()

            def callback(result, error):
                if error:
                    future.set_exception(error)
                elif result is None:
                    # Iteration complete
                    future.set_result(names)
                else:
                    names.append(result.name)

            cursor.each(callback=callback)
            await future
            assert sorted(names) == ['student-%s' % i for i in range(6, 10)]

            # Make sure this kind of notation doesn't create new cursor
            cursor = Student.find()
            cursor_limit = cursor.limit(5)
            cursor_skip = cursor.skip(6)
            assert cursor is cursor_limit is cursor_skip

            # Test clone&rewind as well
            cursor = Student.find()
            cursor2 = cursor.clone()
            await cursor.fetch_next
            await cursor2.fetch_next
            cursor_student = cursor.next_object()
            cursor2_student = cursor2.next_object()
            assert cursor_student == cursor2_student

            # Filter + projection
            cursor = Student.find({'name': 'student-0'}, ['name'])
            students = list(await cursor.to_list(length=100))
            assert len(students) == 1
            assert students[0].name == 'student-0'

        loop.run_until_complete(do_test())

    def test_classroom(self, loop, classroom_model):

        async def do_test():

            student = classroom_model.Student(name='Marty McFly', birthday=datetime(1968, 6, 9))
            await student.commit()
            teacher = classroom_model.Teacher(name='M. Strickland')
            await teacher.commit()
            course = classroom_model.Course(name='Hoverboard 101', teacher=teacher)
            await course.commit()
            assert student.courses is None
            student.courses = []
            assert student.courses == []
            student.courses.append(course)
            await student.commit()
            assert student.to_mongo() == {
                '_id': student.pk,
                'name': 'Marty McFly',
                'birthday': datetime(1968, 6, 9),
                'courses': [course.pk]
            }

        loop.run_until_complete(do_test())

    def test_validation_on_commit(self, loop, instance):

        async def do_test():

            async def io_validate(field, value):
                raise exceptions.ValidationError('Ho boys !')

            @instance.register
            class Dummy(Document):
                required_name = fields.StrField(required=True)
                always_io_fail = fields.IntField(io_validate=io_validate)

            with pytest.raises(exceptions.ValidationError) as exc:
                await Dummy().commit()
            assert exc.value.messages == {'required_name': ['Missing data for required field.']}
            with pytest.raises(exceptions.ValidationError) as exc:
                await Dummy(required_name='required', always_io_fail=42).commit()
            assert exc.value.messages == {'always_io_fail': ['Ho boys !']}

            dummy = Dummy(required_name='required')
            await dummy.commit()
            del dummy.required_name
            with pytest.raises(exceptions.ValidationError) as exc:
                await dummy.commit()
            assert exc.value.messages == {'required_name': ['Missing data for required field.']}

        loop.run_until_complete(do_test())

    def test_reference(self, loop, classroom_model):

        async def do_test():

            teacher = classroom_model.Teacher(name='M. Strickland')
            await teacher.commit()
            course = classroom_model.Course(name='Hoverboard 101', teacher=teacher)
            await course.commit()
            assert isinstance(course.teacher, Reference)
            teacher_fetched = await course.teacher.fetch()
            assert teacher_fetched == teacher
            # Change in referenced document is not seen until referenced
            # document is committed and referencer is reloaded
            teacher.name = 'Dr. Brown'
            teacher_fetched = await course.teacher.fetch()
            assert teacher_fetched.name == 'M. Strickland'
            await teacher.commit()
            teacher_fetched = await course.teacher.fetch()
            assert teacher_fetched.name == 'M. Strickland'
            await course.reload()
            teacher_fetched = await course.teacher.fetch()
            assert teacher_fetched.name == 'Dr. Brown'
            # But we can force reload as soon as referenced document is committed
            # without having to reload the whole referencer
            teacher.name = 'M. Strickland'
            teacher_fetched = await course.teacher.fetch()
            assert teacher_fetched.name == 'Dr. Brown'
            teacher_fetched = await course.teacher.fetch(force_reload=True)
            assert teacher_fetched.name == 'Dr. Brown'
            await teacher.commit()
            teacher_fetched = await course.teacher.fetch()
            assert teacher_fetched.name == 'Dr. Brown'
            teacher_fetched = await course.teacher.fetch(force_reload=True)
            assert teacher_fetched.name == 'M. Strickland'
            # Test bad ref as well
            course.teacher = Reference(classroom_model.Teacher, ObjectId())
            with pytest.raises(exceptions.ValidationError) as exc:
                await course.io_validate()
            assert exc.value.messages == {'teacher': ['Reference not found for document Teacher.']}

        loop.run_until_complete(do_test())

    def test_io_validate(self, loop, instance, classroom_model):

        async def do_test():

            Student = classroom_model.Student

            io_field_value = 'io?'
            io_validate_called = False

            async def io_validate(field, value):
                assert field == IOStudent.schema.fields['io_field']
                assert value == io_field_value
                nonlocal io_validate_called
                io_validate_called = True

            @instance.register
            class IOStudent(Student):
                io_field = fields.StrField(io_validate=io_validate)

            student = IOStudent(name='Marty', io_field=io_field_value)
            assert not io_validate_called

            await student.io_validate()
            assert io_validate_called

        loop.run_until_complete(do_test())

    def test_io_validate_error(self, loop, instance, classroom_model):

        async def do_test():

            Student = classroom_model.Student

            async def io_validate(field, value):
                raise exceptions.ValidationError('Ho boys !')

            @instance.register
            class EmbeddedDoc(EmbeddedDocument):
                io_field = fields.IntField(io_validate=io_validate)

            @instance.register
            class IOStudent(Student):
                io_field = fields.StrField(io_validate=io_validate)
                list_io_field = fields.ListField(fields.IntField(io_validate=io_validate))
                reference_io_field = fields.ReferenceField(classroom_model.Course, io_validate=io_validate)
                embedded_io_field = fields.EmbeddedField(EmbeddedDoc, io_validate=io_validate)

            bad_reference = ObjectId()
            student = IOStudent(
                name='Marty',
                io_field='io?',
                list_io_field=[1, 2],
                reference_io_field=bad_reference,
                embedded_io_field={'io_field': 42}
            )
            with pytest.raises(exceptions.ValidationError) as exc:
                await student.io_validate()
            assert exc.value.messages == {
                'io_field': ['Ho boys !'],
                'list_io_field': {0: ['Ho boys !'], 1: ['Ho boys !']},
                'reference_io_field': ['Ho boys !', 'Reference not found for document Course.'],
                'embedded_io_field': {'io_field': ['Ho boys !']}
            }

        loop.run_until_complete(do_test())

    def test_io_validate_multi_validate(self, loop, instance, classroom_model):

        async def do_test():

            Student = classroom_model.Student
            called = []

            future1 = asyncio.Future()
            future2 = asyncio.Future()
            future3 = asyncio.Future()
            future4 = asyncio.Future()

            async def io_validate11(field, value):
                called.append(1)
                future1.set_result(None)
                await future3
                called.append(4)
                future4.set_result(None)

            async def io_validate12(field, value):
                await future4
                called.append(5)

            async def io_validate21(field, value):
                await future2
                called.append(3)
                future3.set_result(None)

            async def io_validate22(field, value):
                await future1
                called.append(2)
                future2.set_result(None)

            @instance.register
            class IOStudent(Student):
                io_field1 = fields.StrField(io_validate=(io_validate11, io_validate12))
                io_field2 = fields.StrField(io_validate=(io_validate21, io_validate22))

            student = IOStudent(name='Marty', io_field1='io1', io_field2='io2')
            await student.io_validate()
            assert called == [1, 2, 3, 4, 5]

        loop.run_until_complete(do_test())

    def test_io_validate_list(self, loop, instance, classroom_model):

        async def do_test():

            Student = classroom_model.Student
            called = []

            values = [1, 2, 3, 4]
            total = len(values)
            futures = [asyncio.Future() for _ in range(total * 2)]

            async def io_validate(field, value):
                futures[total - value + 1].set_result(None)
                await futures[value]
                called.append(value)

            @instance.register
            class IOStudent(Student):
                io_field = fields.ListField(fields.IntField(io_validate=io_validate))

            student = IOStudent(name='Marty', io_field=values)
            await student.io_validate()
            assert set(called) == set(values)

        loop.run_until_complete(do_test())

    def test_indexes(self, loop, instance):

        async def do_test():

            @instance.register
            class SimpleIndexDoc(Document):
                indexed = fields.StrField()
                no_indexed = fields.IntField()

                class Meta:
                    indexes = ['indexed']

            await SimpleIndexDoc.collection.drop_indexes()

            # Now ask for indexes building
            await SimpleIndexDoc.ensure_indexes()
            indexes = await SimpleIndexDoc.collection.index_information()
            expected_indexes = {
                '_id_': {
                    'key': [('_id', 1)],
                },
                'indexed_1': {
                    'key': [('indexed', 1)],
                }
            }
            assert _stripped(indexes) == expected_indexes

            # Redoing indexes building should do nothing
            await SimpleIndexDoc.ensure_indexes()
            indexes = await SimpleIndexDoc.collection.index_information()
            assert _stripped(indexes) == expected_indexes

        loop.run_until_complete(do_test())

    def test_indexes_inheritance(self, loop, instance):

        async def do_test():

            @instance.register
            class SimpleIndexDoc(Document):
                indexed = fields.StrField()
                no_indexed = fields.IntField()

                class Meta:
                    indexes = ['indexed']

            await SimpleIndexDoc.collection.drop_indexes()

            # Now ask for indexes building
            await SimpleIndexDoc.ensure_indexes()
            indexes = await SimpleIndexDoc.collection.index_information()
            expected_indexes = {
                '_id_': {
                    'key': [('_id', 1)],
                },
                'indexed_1': {
                    'key': [('indexed', 1)],
                }
            }
            assert _stripped(indexes) == expected_indexes

            # Redoing indexes building should do nothing
            await SimpleIndexDoc.ensure_indexes()
            indexes = await SimpleIndexDoc.collection.index_information()
            assert _stripped(indexes) == expected_indexes

        loop.run_until_complete(do_test())

    def test_unique_index(self, loop, instance):

        async def do_test():

            @instance.register
            class UniqueIndexDoc(Document):
                not_unique = fields.StrField(unique=False)
                sparse_unique = fields.IntField(unique=True)
                required_unique = fields.IntField(unique=True, required=True)

            await UniqueIndexDoc.collection.drop()
            await UniqueIndexDoc.collection.drop_indexes()

            # Now ask for indexes building
            await UniqueIndexDoc.ensure_indexes()
            indexes = await UniqueIndexDoc.collection.index_information()
            expected_indexes = {
                '_id_': {
                    'key': [('_id', 1)],
                },
                'required_unique_1': {
                    'key': [('required_unique', 1)],
                    'unique': True
                },
                'sparse_unique_1': {
                    'key': [('sparse_unique', 1)],
                    'unique': True,
                    'sparse': True
                }
            }
            assert _stripped(indexes) == expected_indexes

            # Redoing indexes building should do nothing
            await UniqueIndexDoc.ensure_indexes()
            indexes = await UniqueIndexDoc.collection.index_information()
            assert _stripped(indexes) == expected_indexes

            await UniqueIndexDoc(not_unique='a', required_unique=1).commit()
            await UniqueIndexDoc(not_unique='a', sparse_unique=1, required_unique=2).commit()
            with pytest.raises(exceptions.ValidationError) as exc:
                await UniqueIndexDoc(not_unique='a', required_unique=1).commit()
            assert exc.value.messages == {'required_unique': 'Field value must be unique.'}
            with pytest.raises(exceptions.ValidationError) as exc:
                await UniqueIndexDoc(not_unique='a', sparse_unique=1, required_unique=3).commit()
            assert exc.value.messages == {'sparse_unique': 'Field value must be unique.'}

        loop.run_until_complete(do_test())

    def test_unique_index_compound(self, loop, instance):

        async def do_test():

            @instance.register
            class UniqueIndexCompoundDoc(Document):
                compound1 = fields.IntField()
                compound2 = fields.IntField()
                not_unique = fields.StrField()

                class Meta:
                    # Must define custom index to do that
                    indexes = [{'key': ('compound1', 'compound2'), 'unique': True}]

            await UniqueIndexCompoundDoc.collection.drop()
            await UniqueIndexCompoundDoc.collection.drop_indexes()

            # Now ask for indexes building
            await UniqueIndexCompoundDoc.ensure_indexes()
            indexes = await UniqueIndexCompoundDoc.collection.index_information()
            # Must sort compound indexes to avoid random inconsistence
            indexes['compound1_1_compound2_1']['key'] = sorted(indexes['compound1_1_compound2_1']['key'])
            expected_indexes = {
                '_id_': {
                    'key': [('_id', 1)],
                },
                'compound1_1_compound2_1': {
                    'key': [('compound1', 1), ('compound2', 1)],
                    'unique': True
                }
            }
            assert _stripped(indexes) == expected_indexes

            # Redoing indexes building should do nothing
            await UniqueIndexCompoundDoc.ensure_indexes()
            indexes = await UniqueIndexCompoundDoc.collection.index_information()
            # Must sort compound indexes to avoid random inconsistence
            indexes['compound1_1_compound2_1']['key'] = sorted(indexes['compound1_1_compound2_1']['key'])
            assert _stripped(indexes) == expected_indexes

            # Index is on the tuple (compound1, compound2)
            await UniqueIndexCompoundDoc(not_unique='a', compound1=1, compound2=1).commit()
            await UniqueIndexCompoundDoc(not_unique='a', compound1=1, compound2=2).commit()
            await UniqueIndexCompoundDoc(not_unique='a', compound1=2, compound2=1).commit()
            await UniqueIndexCompoundDoc(not_unique='a', compound1=2, compound2=2).commit()
            with pytest.raises(exceptions.ValidationError) as exc:
                await UniqueIndexCompoundDoc(not_unique='a', compound1=1, compound2=1).commit()
            assert exc.value.messages == {
                'compound2': "Values of fields ['compound1', 'compound2'] must be unique together.",
                'compound1': "Values of fields ['compound1', 'compound2'] must be unique together."
            }
            with pytest.raises(exceptions.ValidationError) as exc:
                await UniqueIndexCompoundDoc(not_unique='a', compound1=2, compound2=1).commit()
            assert exc.value.messages == {
                'compound2': "Values of fields ['compound1', 'compound2'] must be unique together.",
                'compound1': "Values of fields ['compound1', 'compound2'] must be unique together."
            }

        loop.run_until_complete(do_test())

    @pytest.mark.xfail
    def test_unique_index_inheritance(self, loop, instance):

        async def do_test():

            @instance.register
            class UniqueIndexParentDoc(Document):
                not_unique = fields.StrField(unique=False)
                unique = fields.IntField(unique=True)

                class Meta:
                    allow_inheritance = True

            @instance.register
            class UniqueIndexChildDoc(UniqueIndexParentDoc):
                child_not_unique = fields.StrField(unique=False)
                child_unique = fields.IntField(unique=True)
                manual_index = fields.IntField()

                class Meta:
                    indexes = ['manual_index']

            await UniqueIndexChildDoc.collection.drop_indexes()

            # Now ask for indexes building
            await UniqueIndexChildDoc.ensure_indexes()
            indexes = [e async for e in UniqueIndexChildDoc.collection.list_indexes()]
            expected_indexes = [
                {
                    'key': {'_id': 1},
                    'name': '_id_',
                    'ns': '%s.unique_index_inheritance_doc' % TEST_DB,
                },
                {
                    'key': {'unique': 1},
                    'name': 'unique_1',
                    'unique': True,
                    'ns': '%s.unique_index_inheritance_doc' % TEST_DB
                },
                {
                    'key': {'manual_index': 1, '_cls': 1},
                    'name': 'manual_index_1__cls_1',
                    'ns': '%s.unique_index_inheritance_doc' % TEST_DB
                },
                {
                    'key': {'_cls': 1},
                    'name': '_cls_1',
                    'unique': True,
                    'ns': '%s.unique_index_inheritance_doc' % TEST_DB
                },
                {
                    'key': {'child_unique': 1, '_cls': 1},
                    'name': 'child_unique_1__cls_1',
                    'unique': True,
                    'ns': '%s.unique_index_inheritance_doc' % TEST_DB
                }
            ]
            assert name_sorted(indexes) == name_sorted(expected_indexes)

            # Redoing indexes building should do nothing
            await UniqueIndexChildDoc.ensure_indexes()
            indexes = [e for e in (await UniqueIndexChildDoc.collection.list_indexes())]
            assert name_sorted(indexes) == name_sorted(expected_indexes)

        loop.run_until_complete(do_test())

    def test_inheritance_search(self, loop, instance):

        async def do_test():

            @instance.register
            class InheritanceSearchParent(Document):
                pf = fields.IntField()

                class Meta:
                    allow_inheritance = True

            @instance.register
            class InheritanceSearchChild1(InheritanceSearchParent):
                c1f = fields.IntField()

                class Meta:
                    allow_inheritance = True

            @instance.register
            class InheritanceSearchChild1Child(InheritanceSearchChild1):
                sc1f = fields.IntField()

            @instance.register
            class InheritanceSearchChild2(InheritanceSearchParent):
                c2f = fields.IntField(required=True)

            await InheritanceSearchParent.collection.drop()

            await InheritanceSearchParent(pf=0).commit()
            await InheritanceSearchChild1(pf=1, c1f=1).commit()
            await InheritanceSearchChild1Child(pf=1, sc1f=1).commit()
            await InheritanceSearchChild2(pf=2, c2f=2).commit()

            assert (await InheritanceSearchParent.count_documents()) == 4
            assert (await InheritanceSearchChild1.count_documents()) == 2
            assert (await InheritanceSearchChild1Child.count_documents()) == 1
            assert (await InheritanceSearchChild2.count_documents()) == 1

            res = await InheritanceSearchParent.find_one({'sc1f': 1})
            assert isinstance(res, InheritanceSearchChild1Child)

            cursor = InheritanceSearchParent.find({'pf': 1})
            for r in (await cursor.to_list(length=100)):
                assert isinstance(r, InheritanceSearchChild1)

            isc = InheritanceSearchChild1(pf=2, c1f=2)
            await isc.commit()
            res = await InheritanceSearchChild1.find_one(isc.id)
            assert res == isc

            res = await InheritanceSearchChild1.find_one(isc.id, ['c1f'])
            assert res.c1f == 2

        loop.run_until_complete(do_test())

    def test_search(self, loop, instance):

        async def do_test():

            @instance.register
            class Author(EmbeddedDocument):
                name = fields.StrField(attribute='an')

            @instance.register
            class Chapter(EmbeddedDocument):
                name = fields.StrField(attribute='cn')

            @instance.register
            class Book(Document):
                title = fields.StrField(attribute='t')
                author = fields.EmbeddedField(Author, attribute='a')
                chapters = fields.ListField(fields.EmbeddedField(Chapter), attribute='c')

            await Book.collection.drop()
            await Book(
                title='The Hobbit',
                author={'name': 'JRR Tolkien'},
                chapters=[
                    {'name': 'An Unexpected Party'},
                    {'name': 'Roast Mutton'},
                    {'name': 'A Short Rest'},
                    {'name': 'Over Hill And Under Hill'},
                    {'name': 'Riddles In The Dark'}
                ]
            ).commit()
            await Book(
                title="Harry Potter and the Philosopher's Stone",
                author={'name': 'JK Rowling'},
                chapters=[
                    {'name': 'The Boy Who Lived'},
                    {'name': 'The Vanishing Glass'},
                    {'name': 'The Letters from No One'},
                    {'name': 'The Keeper of the Keys'},
                    {'name': 'Diagon Alley'}
                ]
            ).commit()
            await Book(
                title='A Game of Thrones',
                author={'name': 'George RR Martin'},
                chapters=[
                    {'name': 'Prologue'},
                    {'name': 'Bran I'},
                    {'name': 'Catelyn I'},
                    {'name': 'Daenerys I'},
                    {'name': 'Eddard I'},
                    {'name': 'Jon I'}
                ]
            ).commit()

            res = await Book.count_documents({'title': 'The Hobbit'})
            assert res == 1
            res = await Book.count_documents({'author.name': {'$in': ['JK Rowling', 'JRR Tolkien']}})
            assert res == 2
            res = await Book.count_documents({'$and': [{'chapters.name': 'Roast Mutton'}, {'title': 'The Hobbit'}]})
            assert res == 1
            res = await Book.count_documents({'chapters.name': {'$all': ['Roast Mutton', 'A Short Rest']}})
            assert res == 1

        loop.run_until_complete(do_test())

    def test_pre_post_hooks(self, loop, instance):

        async def do_test():

            callbacks = []

            @instance.register
            class Person(Document):
                name = fields.StrField()
                age = fields.IntField()

                def pre_insert(self):
                    callbacks.append('pre_insert')

                def pre_update(self):
                    callbacks.append('pre_update')

                def pre_delete(self):
                    callbacks.append('pre_delete')

                def post_insert(self, ret):
                    assert isinstance(ret, InsertOneResult)
                    callbacks.append('post_insert')

                def post_update(self, ret):
                    assert isinstance(ret, UpdateResult)
                    callbacks.append('post_update')

                def post_delete(self, ret):
                    assert isinstance(ret, DeleteResult)
                    callbacks.append('post_delete')

            p = Person(name='John', age=20)
            await p.commit()
            assert callbacks == ['pre_insert', 'post_insert']

            callbacks.clear()
            p.age = 22
            await p.commit({'age': 22})
            assert callbacks == ['pre_update', 'post_update']

            callbacks.clear()
            await p.delete()
            assert callbacks == ['pre_delete', 'post_delete']

        loop.run_until_complete(do_test())

    def test_pre_post_hooks_with_defers(self, loop, instance):

        async def do_test():

            events = []

            @instance.register
            class Person(Document):
                name = fields.StrField()
                age = fields.IntField()

                async def pre_insert(self):
                    events.append('start pre_insert')
                    future = asyncio.Future()
                    future.set_result(True)
                    await future
                    events.append('end pre_insert')

                async def post_insert(self, ret):
                    events.append('start post_insert')
                    future = asyncio.Future()
                    future.set_result(True)
                    await future
                    events.append('end post_insert')

            p = Person(name='John', age=20)
            await p.commit()
            assert events == [
                'start pre_insert',
                'end pre_insert',
                'start post_insert',
                'end post_insert'
            ]

        loop.run_until_complete(do_test())

    def test_modify_in_pre_hook(self, loop, instance):

        async def do_test():

            @instance.register
            class Person(Document):
                version = fields.IntField(required=True, attribute='_version')
                name = fields.StrField()
                age = fields.IntField()

                def pre_insert(self):
                    self.version = 1

                def pre_update(self):
                    # Prevent concurrency by checking a version number on update
                    last_version = self.version
                    self.version += 1
                    return {'version': last_version}

                def pre_delete(self):
                    return {'version': self.version}

            p = Person(name='John', age=20)
            await p.commit()

            assert p.version == 1
            p_concurrent = await Person.find_one(p.pk)

            p.age = 22
            await p.commit()
            assert p.version == 2

            # Concurrent should not be able to commit it modifications
            p_concurrent.name = 'John'
            with pytest.raises(exceptions.UpdateError):
                await p_concurrent.commit()

            await p_concurrent.reload()
            assert p_concurrent.version == 2

            p.age = 24
            await p.commit()
            assert p.version == 3
            await p.delete()
            await p.commit()
            with pytest.raises(exceptions.DeleteError):
                await p_concurrent.delete()
            await p.delete()

        loop.run_until_complete(do_test())



import importlib

import pytest

from bson import ObjectId

from umongo import Document, fields, EmbeddedDocument
from umongo.instance import Instance
from umongo.document import DocumentTemplate, DocumentImplementation
from umongo.embedded_document import EmbeddedDocumentTemplate, EmbeddedDocumentImplementation
from umongo.frameworks import (
    MongoMockInstance, MotorAsyncIOInstance, TxMongoInstance, PyMongoInstance
)
from umongo.exceptions import (
    AlreadyRegisteredDocumentError, NotRegisteredDocumentError, NoDBDefinedError
)

from .common import MockedDB, MockedInstance


# Try to retrieve framework's db to test against each of them
DB_AND_INSTANCE_PER_FRAMEWORK = [(MockedDB('my_db'), MockedInstance)]
for name, inst in (('mongomock', MongoMockInstance),
                   ('motor_asyncio', MotorAsyncIOInstance),
                   ('txmongo', TxMongoInstance),
                   ('pymongo', PyMongoInstance)):
    mod = importlib.import_module('tests.frameworks.test_%s' % name)
    if not mod.dep_error:
        DB_AND_INSTANCE_PER_FRAMEWORK.append((mod.make_db(), inst))


@pytest.fixture(params=DB_AND_INSTANCE_PER_FRAMEWORK)
def db(request):
    return request.param[0]


@pytest.fixture(params=DB_AND_INSTANCE_PER_FRAMEWORK)
def db_and_instance(request):
    return request.param


class TestInstance:

    def test_already_register(self, instance):

        class Doc(Document):
            pass

        implementation = instance.register(Doc)
        assert issubclass(implementation, DocumentImplementation)
        with pytest.raises(AlreadyRegisteredDocumentError):
            instance.register(Doc)

        class Embedded(EmbeddedDocument):
            pass

        implementation = instance.register(Embedded)
        assert issubclass(implementation, EmbeddedDocumentImplementation)
        with pytest.raises(AlreadyRegisteredDocumentError):
            instance.register(Embedded)

    def test_not_register_documents(self, instance):

        with pytest.raises(NotRegisteredDocumentError):
            @instance.register
            class Doc1(Document):
                ref = fields.ReferenceField('DummyDoc')
            Doc1(ref=ObjectId('56dee8dd1d41c8860b263d86'))

        with pytest.raises(NotRegisteredDocumentError):
            @instance.register
            class Doc2(Document):
                nested = fields.EmbeddedField('DummyNested')
            Doc2(nested={})

    def test_multiple_instances(self, db):
        instance1 = Instance(db)
        instance2 = Instance(db)

        class Doc(Document):
            pass

        class Embedded(EmbeddedDocument):
            pass

        Doc1 = instance1.register(Doc)
        Doc2 = instance2.register(Doc)
        Embedded1 = instance1.register(Embedded)
        Embedded2 = instance2.register(Embedded)

        assert issubclass(Doc1, DocumentImplementation)
        assert issubclass(Doc2, DocumentImplementation)
        assert issubclass(Embedded1, EmbeddedDocumentImplementation)
        assert issubclass(Embedded2, EmbeddedDocumentImplementation)
        assert Doc1.opts.instance is instance1
        assert Doc2.opts.instance is instance2
        assert Embedded1.opts.instance is instance1
        assert Embedded2.opts.instance is instance2

    def test_register_other_implementation(self, db):
        instance1 = Instance(db)
        instance2 = Instance(db)

        class Doc(Document):
            pass

        doc_instance1_cls = instance1.register(Doc)
        doc_instance2_cls = instance2.register(doc_instance1_cls)
        assert issubclass(doc_instance2_cls, DocumentImplementation)
        with pytest.raises(AlreadyRegisteredDocumentError):
            instance2.register(Doc)

        class Embedded(EmbeddedDocument):
            pass

        embedded_instance1_cls = instance1.register(Embedded)
        embedded_instance2_cls = instance2.register(embedded_instance1_cls)
        assert issubclass(embedded_instance2_cls, EmbeddedDocumentImplementation)
        with pytest.raises(AlreadyRegisteredDocumentError):
            instance2.register(Embedded)

    def test_parent_not_registered(self, instance):
        class Parent(Document):
            pass

        with pytest.raises(NotRegisteredDocumentError):
            @instance.register
            class Child(Parent):
                pass

        class ParentEmbedded(EmbeddedDocument):
            pass

        with pytest.raises(NotRegisteredDocumentError):
            @instance.register
            class ChildEmbedded(ParentEmbedded):
                pass

    def test_retrieve_registered(self, instance):
        class Doc(Document):
            pass

        class Embedded(EmbeddedDocument):
            pass

        Doc_imp = instance.register(Doc)
        Embedded_imp = instance.register(Embedded)

        assert instance.retrieve_document('Doc') is Doc_imp
        assert instance.retrieve_document(Doc) is Doc_imp
        assert instance.retrieve_embedded_document('Embedded') is Embedded_imp
        assert instance.retrieve_embedded_document(Embedded) is Embedded_imp

        with pytest.raises(NotRegisteredDocumentError):
            instance.retrieve_document('Dummy')

        with pytest.raises(NotRegisteredDocumentError):
            instance.retrieve_embedded_document('Dummy')

    def test_mix_doc_and_embedded(self, instance):
        @instance.register
        class Doc(Document):
            pass

        @instance.register
        class Embedded(EmbeddedDocument):
            pass

        with pytest.raises(NotRegisteredDocumentError):
            instance.retrieve_document('Embedded')

        with pytest.raises(NotRegisteredDocumentError):
            instance.retrieve_embedded_document('Doc')

    def test_lazy_loader_instance(self, db_and_instance):
        db, lazy_instance = db_and_instance
        instance = lazy_instance()

        class Doc(Document):
            pass

        doc_impl_cls = instance.register(Doc)

        with pytest.raises(NoDBDefinedError):
            doc_impl_cls.collection

        instance.init(db)

        assert doc_impl_cls.collection == db['doc']

    def test_patched_fields(self, db):

        instance1 = Instance(db)
        instance2 = Instance(db)

        class Embedded(EmbeddedDocument):
            simple = fields.IntField()

        Embedded1 = instance1.register(Embedded)
        Embedded2 = instance2.register(Embedded1)

        class ToRef(Document):
            pass

        ToRef1 = instance1.register(ToRef)
        instance2.register(ToRef)

        class Doc(Document):
            embedded = fields.EmbeddedField(Embedded1)
            ref = fields.ReferenceField(ToRef1)

        Doc1 = instance1.register(Doc)
        Doc2 = instance2.register(Doc)

        assert issubclass(Doc.embedded.embedded_document, EmbeddedDocumentTemplate)
        assert issubclass(
            Doc1.schema.fields['embedded'].embedded_document, EmbeddedDocumentTemplate)
        assert issubclass(
            Doc2.schema.fields['embedded'].embedded_document, EmbeddedDocumentTemplate)
        assert issubclass(
            Doc1.schema.fields['embedded'].embedded_document_cls, EmbeddedDocumentImplementation)
        assert issubclass(
            Doc2.schema.fields['embedded'].embedded_document_cls, EmbeddedDocumentImplementation)

        assert issubclass(Doc.ref.document, DocumentTemplate)
        assert issubclass(Doc1.schema.fields['ref'].document, DocumentTemplate)
        assert issubclass(Doc2.schema.fields['ref'].document, DocumentTemplate)
        assert issubclass(Doc1.schema.fields['ref'].document_cls, DocumentImplementation)
        assert issubclass(Doc2.schema.fields['ref'].document_cls, DocumentImplementation)

        assert Embedded1.schema.fields['simple'].instance is instance1
        assert Embedded1.opts.instance is instance1
        assert Embedded2.schema.fields['simple'].instance is instance2
        assert Embedded2.opts.instance is instance2

        assert Doc1.schema.fields['embedded'].instance is instance1
        assert Doc1.opts.instance is instance1
        assert Doc2.schema.fields['embedded'].instance is instance2
        assert Doc2.opts.instance is instance2



import pymongo

from umongo.document import DocumentImplementation
from umongo.instance import Instance, LazyLoaderInstance
from umongo.builder import BaseBuilder
from umongo.frameworks import register_builder


TEST_DB = 'umongo_test'


# Use a sync driver for easily drop the test database
con = pymongo.MongoClient()


# Provide mocked database, collection and builder for easier testing


class MockedCollection():

    def __init__(self, db, name):
        self.db = db
        self.name = name

    def __eq__(self, other):
        return (isinstance(other, MockedCollection) and
                self.db == other.db and self.name == other.name)

    def __repr__(self):
        return "<%s db=%s, name=%s>" % (self.__class__.__name__, self.db, self.name)


class MockedDB:

    def __init__(self, name):
        self.name = name
        self.cols = {}

    def __getattr__(self, name):
        if name not in self.cols:
            self.cols[name] = MockedCollection(self, name)
        return self.cols[name]

    def __getitem__(self, name):
        if name not in self.cols:
            self.cols[name] = MockedCollection(self, name)
        return self.cols[name]

    def __eq__(self, other):
        return isinstance(other, MockedDB) and self.name == other.name

    def __repr__(self):
        return "<%s name=%s>" % (self.__class__.__name__, self.name)


class MockedBuilder(BaseBuilder):

    BASE_DOCUMENT_CLS = DocumentImplementation

    @staticmethod
    def is_compatible_with(db):
        return isinstance(db, MockedDB)


register_builder(MockedBuilder)


class MockedInstance(LazyLoaderInstance):
    BUILDER_CLS = MockedBuilder


class BaseTest:

    def setup(self):
        self.instance = Instance(MockedDB('my_moked_db'))


class BaseDBTest:

    def setup(self):
        con.drop_database(TEST_DB)



from .exceptions import (
    NotRegisteredDocumentError, AlreadyRegisteredDocumentError, NoDBDefinedError)
from .document import DocumentTemplate
from .template import get_template


class BaseInstance:
    """
    Base class for instance.

    Instances aims at collecting and implementing :class:`umongo.template.Template`::

        # Doc is a template, cannot use it for the moment
        class Doc(DocumentTemplate):
            pass

        instance = Instance()
        # doc_cls is the instance's implementation of Doc
        doc_cls = instance.register(Doc)
        # Implementations are registered as attribute into the instance
        instance.Doc is doc_cls
        # Now we can work with the implementations
        doc_cls.find()

    .. note::
        Instance registration is divided between :class:`umongo.Document` and
        :class:`umongo.EmbeddedDocument`.
    """

    BUILDER_CLS = None

    def __init__(self, templates=()):
        assert self.BUILDER_CLS, 'BUILDER_CLS must be defined.'
        self.builder = self.BUILDER_CLS(self)
        self._doc_lookup = {}
        self._embedded_lookup = {}
        for template in templates:
            self.register(template)

    @property
    def db(self):
        """Database used within the instance."""
        raise NotImplementedError

    def retrieve_document(self, name_or_template):
        """
        Retrieve a :class:`umongo.document.DocumentImplementation` registered into this
        instance from it name or it template class (i.e. :class:`umongo.Document`).
        """
        if not isinstance(name_or_template, str):
            name_or_template = name_or_template.__name__
        if name_or_template not in self._doc_lookup:
            raise NotRegisteredDocumentError(
                'Unknown document class `%s`' % name_or_template)
        return self._doc_lookup[name_or_template]

    def retrieve_embedded_document(self, name_or_template):
        """
        Retrieve a :class:`umongo.embedded_document.EmbeddedDocumentImplementation`
        registered into this instance from it name or it template class
        (i.e. :class:`umongo.EmbeddedDocument`).
        """
        if not isinstance(name_or_template, str):
            name_or_template = name_or_template.__name__
        if name_or_template not in self._embedded_lookup:
            raise NotRegisteredDocumentError(
                'Unknown embedded document class `%s`' % name_or_template)
        return self._embedded_lookup[name_or_template]

    def register(self, template, as_attribute=True):
        """
        Generate an :class:`umongo.template.Implementation` from the given
        :class:`umongo.template.Template` for this instance.

        :param template: :class:`umongo.template.Template` to implement
        :param as_attribute:
            Make the generated :class:`umongo.template.Implementation` available
            as this instance's attribute.

        :return: The :class:`umongo.template.Implementation` generated

        .. note::
            This method can be used as a decorator. This is useful when you
            only have a single instance to work with to directly use the
            class you defined::

                @instance.register
                class MyEmbedded(EmbeddedDocument):
                    pass

                @instance.register
                class MyDoc(Document):
                    emb = fields.EmbeddedField(MyEmbedded)

                MyDoc.find()

        """
        # Retrieve the template if another implementation has been provided instead
        template = get_template(template)
        if issubclass(template, DocumentTemplate):
            implementation = self._register_doc(template)
        else:  # EmbeddedDocumentTemplate
            implementation = self._register_embedded_doc(template)
        if as_attribute:
            setattr(self, implementation.__name__, implementation)
        return implementation

    def _register_doc(self, template):
        implementation = self.builder.build_document_from_template(template)
        if hasattr(self, implementation.__name__):
            raise AlreadyRegisteredDocumentError(
                'Document `%s` already registered' % implementation.__name__)
        self._doc_lookup[implementation.__name__] = implementation
        return implementation

    def _register_embedded_doc(self, template):
        implementation = self.builder.build_embedded_document_from_template(template)
        if hasattr(self, implementation.__name__):
            raise AlreadyRegisteredDocumentError(
                'EmbeddedDocument `%s` already registered' % implementation.__name__)
        self._embedded_lookup[implementation.__name__] = implementation
        return implementation


class Instance(BaseInstance):
    """
    Automatically configured instance according to the type of
    the provided database.
    """

    def __init__(self, db, templates=()):
        self._db = db
        # Dynamically find a builder compatible with the db
        from .frameworks import find_builder_from_db
        self.BUILDER_CLS = find_builder_from_db(db)
        super().__init__(templates=templates)

    @property
    def db(self):
        return self._db


class LazyLoaderInstance(BaseInstance):
    """
    Base class for instance with database lazy loading.

    .. note::
        This class should not be used directly but instead overloaded.
        See :class:`umongo.PyMongoInstance` for example.

    """

    def __init__(self, templates=()):
        self._db = None
        super().__init__(templates=templates)

    @property
    def db(self):
        if not self._db:
            raise NoDBDefinedError('init must be called to define a db')
        return self._db

    def init(self, db):
        """
        Set the database to use whithin this instance.

        .. note::
            The documents registered in the instance cannot be used
            before this function is called.
        """
        assert self.BUILDER_CLS.is_compatible_with(db)
        self._db = db



from twisted.internet.defer import (
    inlineCallbacks, Deferred, DeferredList, returnValue, maybeDeferred)
from txmongo import filter as qf
from txmongo.database import Database
from pymongo.errors import DuplicateKeyError

from ..builder import BaseBuilder
from ..document import DocumentImplementation
from ..data_proxy import missing
from ..data_objects import Reference
from ..exceptions import (
    NotCreatedError, UpdateError, DeleteError, ValidationError, NoneReferenceError
)
from ..fields import ReferenceField, ListField, EmbeddedField
from ..query_mapper import map_query

from .tools import cook_find_filter


class TxMongoDocument(DocumentImplementation):

    __slots__ = ()

    opts = DocumentImplementation.opts

    @inlineCallbacks
    def reload(self):
        """
        Retrieve and replace document's data by the ones in database.

        Raises :class:`umongo.exceptions.NotCreatedError` if the document
        doesn't exist in database.
        """
        if not self.is_created:
            raise NotCreatedError("Document doesn't exists in database")
        ret = yield self.collection.find_one(self.pk)
        if ret is None:
            raise NotCreatedError("Document doesn't exists in database")
        self._data = self.DataProxy()
        self._data.from_mongo(ret)

    @inlineCallbacks
    def commit(self, io_validate_all=False, conditions=None):
        """
        Commit the document in database.
        If the document doesn't already exist it will be inserted, otherwise
        it will be updated.

        :param io_validate_all:
        :param conditions: only perform commit if matching record in db
            satisfies condition(s) (e.g. version number).
            Raises :class:`umongo.exceptions.UpdateError` if the
            conditions are not satisfied.
         :return: A :class:`pymongo.results.UpdateResult` or
            :class:`pymongo.results.InsertOneResult` depending of the operation.
       """
        try:
            if self.is_created:
                if self.is_modified():
                    query = conditions or {}
                    query['_id'] = self.pk
                    # pre_update can provide additional query filter and/or
                    # modify the fields' values
                    additional_filter = yield maybeDeferred(self.pre_update)
                    if additional_filter:
                        query.update(map_query(additional_filter, self.schema.fields))
                    self.required_validate()
                    yield self.io_validate(validate_all=io_validate_all)
                    payload = self._data.to_mongo(update=True)
                    ret = yield self.collection.update_one(query, payload)
                    if ret.matched_count != 1:
                        raise UpdateError(ret)
                    yield maybeDeferred(self.post_update, ret)
                else:
                    ret = None
            elif conditions:
                raise RuntimeError('Document must already exist in database to use `conditions`.')
            else:
                yield maybeDeferred(self.pre_insert)
                self.required_validate()
                yield self.io_validate(validate_all=io_validate_all)
                payload = self._data.to_mongo(update=False)
                ret = yield self.collection.insert_one(payload)
                # TODO: check ret ?
                self._data.set(self.pk_field, ret.inserted_id)
                self.is_created = True
                yield maybeDeferred(self.post_insert, ret)
        except DuplicateKeyError as exc:
            # Need to dig into error message to find faulting index
            errmsg = exc.details['errmsg']
            for index in self.opts.indexes:
                if ('.$%s' % index.document['name'] in errmsg or
                        ' %s ' % index.document['name'] in errmsg):
                    keys = index.document['key'].keys()
                    if len(keys) == 1:
                        msg = self.schema.fields[keys[0]].error_messages['unique']
                        raise ValidationError({keys[0]: msg})
                    fields = self.schema.fields
                    # Compound index (sort value to make testing easier)
                    keys = sorted(keys)
                    raise ValidationError(
                        {
                            k: fields[k].error_messages['unique_compound'].format(fields=keys)
                            for k in keys
                        }
                    )
        self._data.clear_modified()
        return ret

    @inlineCallbacks
    def delete(self, conditions=None):
        """
        Remove the document from database.

        :param conditions: Only perform delete if matching record in db
            satisfies condition(s) (e.g. version number).
            Raises :class:`umongo.exceptions.DeleteError` if the
            conditions are not satisfied.
        Raises :class:`umongo.exceptions.NotCreatedError` if the document
        is not created (i.e. ``doc.is_created`` is False)
        Raises :class:`umongo.exceptions.DeleteError` if the document
        doesn't exist in database.

        :return: A :class:`pymongo.results.DeleteResult`
        """
        if not self.is_created:
            raise NotCreatedError("Document doesn't exists in database")
        query = conditions or {}
        query['_id'] = self.pk
        # pre_delete can provide additional query filter
        additional_filter = yield maybeDeferred(self.pre_delete)
        if additional_filter:
            query.update(map_query(additional_filter, self.schema.fields))
        ret = yield self.collection.delete_one(query)
        if ret.deleted_count != 1:
            raise DeleteError(ret)
        self.is_created = False
        yield maybeDeferred(self.post_delete, ret)
        return ret

    def io_validate(self, validate_all=False):
        """
        Run the io_validators of the document's fields.

        :param validate_all: If False only run the io_validators of the
            fields that have been modified.
        """
        if validate_all:
            return _io_validate_data_proxy(self.schema, self._data)
        return _io_validate_data_proxy(
            self.schema, self._data, partial=self._data.get_modified_fields())

    @classmethod
    @inlineCallbacks
    def find_one(cls, filter=None, *args, **kwargs):
        """
        Find a single document in database.
        """
        filter = cook_find_filter(cls, filter)
        ret = yield cls.collection.find_one(filter, *args, **kwargs)
        if ret is not None:
            ret = cls.build_from_mongo(ret, use_cls=True)
        return ret

    @classmethod
    @inlineCallbacks
    def find(cls, filter=None, *args, **kwargs):
        """
        Find a list document in database.

        Returns a list of Documents.
        """
        filter = cook_find_filter(cls, filter)
        raw_cursor_or_list = yield cls.collection.find(filter, *args, **kwargs)
        return [cls.build_from_mongo(e, use_cls=True) for e in raw_cursor_or_list]

    @classmethod
    @inlineCallbacks
    def find_with_cursor(cls, filter=None, *args, **kwargs):
        """
        Find a list document in database.

        Returns a cursor that provides Documents.
        """
        filter = cook_find_filter(cls, filter)
        raw_cursor_or_list = yield cls.collection.find_with_cursor(filter, *args, **kwargs)

        def wrap_raw_results(result):
            cursor = result[1]
            if cursor is not None:
                cursor.addCallback(wrap_raw_results)
            return ([cls.build_from_mongo(e, use_cls=True) for e in result[0]], cursor)

        return wrap_raw_results(raw_cursor_or_list)

    @classmethod
    def count(cls, filter=None, **kwargs):
        """
        Get the number of documents in this collection.
        """
        filter = cook_find_filter(cls, filter)
        return cls.collection.count(filter=filter, **kwargs)

    @classmethod
    @inlineCallbacks
    def ensure_indexes(cls):
        """
        Check&create if needed the Document's indexes in database
        """
        for index in cls.opts.indexes:
            kwargs = index.document.copy()
            keys = kwargs.pop('key')
            index = qf.sort(keys.items())
            yield cls.collection.create_index(index, **kwargs)


def _errback_factory(errors, field=None):

    def errback(err):
        if isinstance(err.value, ValidationError):
            if field is not None:
                errors[field] = err.value.messages
            else:
                errors.extend(err.value.messages)
        else:
            raise err.value

    return errback


# Run multiple validators and collect all errors in one
@inlineCallbacks
def _run_validators(validators, field, value):
    errors = []
    defers = []
    for validator in validators:
        try:
            defer = validator(field, value)
        except ValidationError as exc:
            errors.extend(exc.messages)
        else:
            assert isinstance(defer, Deferred), 'io_validate functions must return a Deferred'
            defer.addErrback(_errback_factory(errors))
            defers.append(defer)
    yield DeferredList(defers)
    if errors:
        raise ValidationError(errors)


@inlineCallbacks
def _io_validate_data_proxy(schema, data_proxy, partial=None):
    errors = {}
    defers = []
    for name, field in schema.fields.items():
        if partial and name not in partial:
            continue
        data_name = field.attribute or name
        value = data_proxy._data[data_name]
        if value is missing:
            continue
        try:
            if field.io_validate_recursive:
                yield field.io_validate_recursive(field, value)
            if field.io_validate:
                defer = _run_validators(field.io_validate, field, value)
                defer.addErrback(_errback_factory(errors, name))
                defers.append(defer)
        except ValidationError as exc:
            errors[name] = exc.messages
    yield DeferredList(defers)
    if errors:
        raise ValidationError(errors)


def _reference_io_validate(field, value):
    return value.fetch(no_data=True)


@inlineCallbacks
def _list_io_validate(field, value):
    validators = field.inner.io_validate
    if not validators or not value:
        return
    errors = {}
    defers = []
    for idx, exc in enumerate(value):
        defer = _run_validators(validators, field.inner, exc)
        defer.addErrback(_errback_factory(errors, idx))
        defers.append(defer)
    yield DeferredList(defers)
    if errors:
        raise ValidationError(errors)


def _embedded_document_io_validate(field, value):
    return _io_validate_data_proxy(value.schema, value._data)


class TxMongoReference(Reference):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._document = None

    @inlineCallbacks
    def fetch(self, no_data=False, force_reload=False):
        if not self._document or force_reload:
            if self.pk is None:
                raise NoneReferenceError('Cannot retrieve a None Reference')
            self._document = yield self.document_cls.find_one(self.pk)
            if not self._document:
                raise ValidationError(self.error_messages['not_found'].format(
                    document=self.document_cls.__name__))
        returnValue(self._document)


class TxMongoBuilder(BaseBuilder):

    BASE_DOCUMENT_CLS = TxMongoDocument

    @staticmethod
    def is_compatible_with(db):
        return isinstance(db, Database)

    def _patch_field(self, field):
        super()._patch_field(field)

        validators = field.io_validate
        if not validators:
            field.io_validate = []
        else:
            if hasattr(validators, '__iter__'):
                validators = list(validators)
            else:
                validators = [validators]
            field.io_validate = validators
        if isinstance(field, ListField):
            field.io_validate_recursive = _list_io_validate
        if isinstance(field, ReferenceField):
            field.io_validate.append(_reference_io_validate)
            field.reference_cls = TxMongoReference
        if isinstance(field, EmbeddedField):
            field.io_validate_recursive = _embedded_document_io_validate



from itertools import zip_longest

import pytest

from umongo import Document, EmbeddedDocument, fields
from umongo.indexes import (
    explicit_key, parse_index,
    IndexModel, ASCENDING, DESCENDING, TEXT, HASHED)

from .common import BaseTest


def assert_indexes(indexes1, indexes2):
    if hasattr(indexes1, '__iter__'):
        for e1, e2 in zip_longest(indexes1, indexes2):
            assert e1, "missing index %s" % e2.document
            assert e2, "too much indexes: %s" % e1.document
            assert e1.document == e2.document
    else:
        assert indexes1.document == indexes2.document


class TestIndexes(BaseTest):

    def test_parse_index(self):
        for value, expected in (
            ('my_index', IndexModel([('my_index', ASCENDING)])),
            ('+my_index', IndexModel([('my_index', ASCENDING)])),
            ('-my_index', IndexModel([('my_index', DESCENDING)])),
            ('$my_index', IndexModel([('my_index', TEXT)])),
            ('#my_index', IndexModel([('my_index', HASHED)])),
            # Compound indexes
            (('index1', '-index2'), IndexModel([('index1', ASCENDING), ('index2', DESCENDING)])),
            # No changes if not needed
            (IndexModel([('my_index', ASCENDING)]), IndexModel([('my_index', ASCENDING)])),
            # Custom index
            (
                {
                    'name': 'my-custom-index',
                    'key': ['+index1', '-index2'],
                    'sparse': True,
                    'unique': True,
                    'expireAfterSeconds': 42
                },
                IndexModel([('index1', ASCENDING), ('index2', DESCENDING)],
                           name='my-custom-index', sparse=True,
                           unique=True, expireAfterSeconds=42)
            ),
        ):
            assert_indexes(parse_index(value), expected)

    def test_explicit_key(self):
        for value, expected in (
            ('my_index', ('my_index', ASCENDING)),
            ('+my_index', ('my_index', ASCENDING)),
            ('-my_index', ('my_index', DESCENDING)),
            ('$my_index', ('my_index', TEXT)),
            ('#my_index', ('my_index', HASHED)),
            # No changes if not needed
            (('my_index', ASCENDING), ('my_index', ASCENDING)),
        ):
            assert explicit_key(value) == expected

    def test_inheritance(self):

        @self.instance.register
        class Parent(Document):
            last_name = fields.StrField()

            class Meta:
                allow_inheritance = True
                indexes = ['last_name']

        @self.instance.register
        class Child(Parent):
            first_name = fields.StrField()

            class Meta:
                indexes = ['-first_name']

        assert_indexes(Parent.opts.indexes, [IndexModel([('last_name', ASCENDING)])])
        assert_indexes(Child.opts.indexes,
            [
                IndexModel([('last_name', ASCENDING)]),
                IndexModel([('first_name', DESCENDING), ('_cls', ASCENDING)]),
                IndexModel([('_cls', ASCENDING)])
            ])

    def test_bad_index(self):
        for bad in [1, None, object()]:
            with pytest.raises(TypeError) as exc:
                parse_index(1)
            assert exc.value.args[0] == (
                'Index type must be <str>, <list>, <dict> or <pymongo.IndexModel>')

    def test_nested_indexes(self):

        @self.instance.register
        class NestedDoc(EmbeddedDocument):
            simple = fields.StrField()
            listed = fields.ListField(fields.StrField())

        @self.instance.register
        class Doc(Document):
            nested = fields.EmbeddedField(NestedDoc)
            listed = fields.ListField(fields.EmbeddedField(NestedDoc))

            class Meta:
                indexes = ['nested', 'nested.simple', 'listed', 'listed.simple', 'listed.listed']

        assert_indexes(Doc.opts.indexes,
            [
                IndexModel([('nested', ASCENDING)]),
                IndexModel([('nested.simple', ASCENDING)]),
                IndexModel([('listed', ASCENDING)]),
                IndexModel([('listed.simple', ASCENDING)]),
                IndexModel([('listed.listed', ASCENDING)]),
            ])



import pytest

from bson import ObjectId
from marshmallow import ValidationError, missing

from umongo.data_proxy import data_proxy_factory, BaseDataProxy, BaseNonStrictDataProxy
from umongo import Schema, fields, EmbeddedDocument, validate, exceptions

from .common import BaseTest


class TestDataProxy(BaseTest):

    def test_repr(self):

        class MySchema(Schema):
            field_a = fields.IntField(attribute='mongo_field_a')
            field_b = fields.StrField()

        MyDataProxy = data_proxy_factory('My', MySchema())
        d = MyDataProxy({'field_a': 1, 'field_b': 'value'})
        assert MyDataProxy.__name__ == 'MyDataProxy'
        repr_d = repr(d)
        assert repr_d.startswith("<MyDataProxy(")
        assert "'field_a': 1" in repr_d
        assert "'field_b': 'value'" in repr_d

    def test_simple(self):

        class MySchema(Schema):
            a = fields.IntField()
            b = fields.IntField()

        MyDataProxy = data_proxy_factory('My', MySchema())
        d = MyDataProxy()
        d.load({'a': 1, 'b': 2})
        assert d.get('a') == 1
        d.set('b', 3)
        assert d.get('b') == 3
        assert d._data == {'a': 1, 'b': 3}
        assert d.dump() == {'a': 1, 'b': 3}
        d.delete('b')
        assert d._data == {'a': 1, 'b': missing}
        assert d.dump() == {'a': 1}

    def test_load(self):

        class MySchema(Schema):
            a = fields.IntField()
            b = fields.IntField(attribute='in_mongo_b')

        MyDataProxy = data_proxy_factory('My', MySchema())
        d = MyDataProxy()
        d.from_mongo({'a': 1, 'in_mongo_b': 2})
        assert d.to_mongo() == {'a': 1, 'in_mongo_b': 2}

        d.set('a', 3)
        assert d.to_mongo(update=True) == {'$set': {'a': 3}}

        d.from_mongo({'a': 4, 'in_mongo_b': 5})
        assert d.to_mongo(update=True) is None
        assert d.to_mongo() == {'a': 4, 'in_mongo_b': 5}

        d2 = MyDataProxy(data={'a': 4, 'b': 5})
        assert d == d2

    def test_modify(self):

        class MySchema(Schema):
            a = fields.IntField()
            b = fields.IntField(attribute='in_mongo_b')

        MyDataProxy = data_proxy_factory('My', MySchema())
        d = MyDataProxy()
        assert d.get_modified_fields() == set()
        d.load({'a': 1, 'b': 2})
        assert d.get_modified_fields() == {'a', 'b'}
        d.from_mongo({'a': 1, 'in_mongo_b': 2})
        assert d.get_modified_fields() == set()
        assert d.to_mongo() == {'a': 1, 'in_mongo_b': 2}
        assert d.to_mongo(update=True) is None
        d.set('a', 3)
        d.delete('b')
        assert d.to_mongo(update=True) == {'$set': {'a': 3}, '$unset': {'in_mongo_b': ''}}
        d.clear_modified()
        assert d.get_modified_fields() == set()
        assert d.to_mongo(update=True) is None
        assert d.to_mongo() == {'a': 3}

    def test_list_field_modify(self):

        class MySchema(Schema):
            a = fields.ListField(fields.IntField())
            b = fields.ListField(fields.IntField(), attribute='in_mongo_b')

        MyDataProxy = data_proxy_factory('My', MySchema())
        d = MyDataProxy()
        assert d.get_modified_fields() == set()
        d.load({'a': [1], 'b': [2, 2]})
        assert d.get_modified_fields() == {'a', 'b'}
        d.from_mongo({'a': [1], 'in_mongo_b': [2, 2]})
        assert d.get_modified_fields() == set()
        assert d.to_mongo() == {'a': [1], 'in_mongo_b': [2, 2]}
        assert d.to_mongo(update=True) is None
        d.set('a', [3, 3, 3])
        d.delete('b')
        assert d.to_mongo(update=True) == {'$set': {'a': [3, 3, 3]}, '$unset': {'in_mongo_b': ''}}
        d.clear_modified()
        assert d.get_modified_fields() == set()
        assert d.to_mongo() == {'a': [3, 3, 3]}
        assert d.to_mongo(update=True) is None
        d.clear_modified()
        d.load({'a': [1], 'b': [2, 2]})
        d._data['a'].append(1)
        d._data['in_mongo_b'].append(2)
        assert d.get_modified_fields() == {'a', 'b'}
        assert d.to_mongo() == {'a': [1, 1], 'in_mongo_b': [2, 2, 2]}
        assert d.to_mongo(update=True) == {'$set': {'a': [1, 1], 'in_mongo_b': [2, 2, 2]}}
        d.clear_modified()
        del d._data['a'][0]
        del d._data['in_mongo_b'][0]
        assert d.get_modified_fields() == {'a', 'b'}
        assert d.to_mongo() == {'a': [1], 'in_mongo_b': [2, 2]}
        assert d.to_mongo(update=True) == {'$set': {'a': [1], 'in_mongo_b': [2, 2]}}
        d.clear_modified()
        d._data['a'].clear()
        d._data['in_mongo_b'].clear()
        assert d.get_modified_fields() == {'a', 'b'}
        assert d.to_mongo() == {'a': [], 'in_mongo_b': []}
        assert d.to_mongo(update=True) == {'$set': {'a': [], 'in_mongo_b': []}}

    def test_complex_field_clear_modified(self):

        @self.instance.register
        class MyEmbedded(EmbeddedDocument):
            aa = fields.IntField()

        class MySchema(Schema):
            # EmbeddedField need instance to retrieve implementation
            a = fields.EmbeddedField(MyEmbedded, instance=self.instance)
            b = fields.ListField(fields.IntField)

        MyDataProxy = data_proxy_factory('My', MySchema())
        d = MyDataProxy()
        d.load({'a': {'aa': 1}, 'b': [2, 3]})
        assert d.to_mongo() == {'a': {'aa': 1}, 'b': [2, 3]}
        d.get('a').aa = 4
        d.get('b').append(5)
        assert d.to_mongo(update=True) == {'$set': {'a': {'aa': 4}, 'b': [2, 3, 5]}}
        d.clear_modified()
        assert d.to_mongo(update=True) is None
        assert not d.get('a').is_modified()
        assert not d.get('b').is_modified()

    def test_set(self):

        class MySchema(Schema):
            a = fields.IntField()
            b = fields.IntField(attribute='in_mongo_b')
            c = fields.StrField(
                allow_none=True,
                validate=validate.Length(min=1, max=5)
            )
            d = fields.StrField()

        MyDataProxy = data_proxy_factory('My', MySchema())
        d = MyDataProxy()
        d.from_mongo({'a': 1, 'in_mongo_b': 2})
        d.set('a', 3)
        assert d.to_mongo() == {'a': 3, 'in_mongo_b': 2}
        assert d.to_mongo(update=True) == {'$set': {'a': 3}}

        d.from_mongo({'a': 1, 'in_mongo_b': 2})
        d.set('b', 3)
        assert d.to_mongo() == {'a': 1, 'in_mongo_b': 3}
        assert d.to_mongo(update=True) == {'$set': {'in_mongo_b': 3}}

        with pytest.raises(KeyError):
            d.set('in_mongo_b', 2)

        d.from_mongo({})
        d.set('c', None)
        assert d.to_mongo() == {'c': None}
        with pytest.raises(ValidationError):
            d.set('c', '123456')
        with pytest.raises(ValidationError):
            d.set('d', None)

    def test_del(self):

        class MySchema(Schema):
            a = fields.IntField()
            b = fields.IntField(attribute='in_mongo_b')

        MyDataProxy = data_proxy_factory('My', MySchema())
        d = MyDataProxy()
        d.from_mongo({'a': 1, 'in_mongo_b': 2})
        d.delete('b')
        assert d.to_mongo() == {'a': 1}
        assert d.to_mongo(update=True) == {'$unset': {'in_mongo_b': ''}}
        d.delete('a')
        assert d.to_mongo(update=True) == {'$unset': {'a': '', 'in_mongo_b': ''}}

        with pytest.raises(KeyError):
            d.delete('in_mongo_b')

    def test_route_naming(self):

        class MySchema(Schema):
            in_front = fields.IntField(attribute='in_mongo')

        MyDataProxy = data_proxy_factory('My', MySchema())
        d = MyDataProxy()
        with pytest.raises(ValidationError):
            d.load({'in_mongo': 42})
        d.load({'in_front': 42})
        with pytest.raises(KeyError):
            d.get('in_mongo')
        assert d.get('in_front') == 42
        d.set('in_front', 24)
        assert d._data == {'in_mongo': 24}
        assert d.get('in_front') == 24
        assert d.dump() == {'in_front': 24}
        assert d.to_mongo() == {'in_mongo': 24}

    def test_from_mongo(self):

        class MySchema(Schema):
            in_front = fields.IntField(attribute='in_mongo')

        MyDataProxy = data_proxy_factory('My', MySchema())
        d = MyDataProxy()
        with pytest.raises(exceptions.UnknownFieldInDBError):
            d.from_mongo({'in_front': 42})
        d.from_mongo({'in_mongo': 42})
        assert d.get('in_front') == 42

    def test_equality(self):

        class MySchema(Schema):
            a = fields.IntField()
            b = fields.IntField(attribute='in_mongo_b')

        MyDataProxy = data_proxy_factory('My', MySchema())
        d1 = MyDataProxy()
        d1.load({'a': 1, 'b': 2})
        assert d1 == {'a': 1, 'in_mongo_b': 2}

        d2 = MyDataProxy()
        d2.load({'a': 1, 'b': 2})
        assert d1 == d2

        assert d1 != None  # noqa: E711 (None comparison)
        assert d1 != missing
        assert None != d1  # noqa: E711 (None comparison)
        assert missing != d1

    def test_share_ressources(self):

        class MySchema(Schema):
            a = fields.IntField()
            b = fields.IntField(attribute='in_mongo_b')

        MyDataProxy = data_proxy_factory('My', MySchema())
        d1 = MyDataProxy()
        d2 = MyDataProxy()
        for field in ('schema', '_fields', '_fields_from_mongo_key'):
            assert getattr(d1, field) is getattr(d2, field)
        d1.load({'a': 1})
        d2.load({'b': 2})
        assert d1 != d2

    def test_set_to_missing_fields(self):

        class MySchema(Schema):
            a = fields.IntField()
            b = fields.IntField(attribute='in_mongo_b')

        MyDataProxy = data_proxy_factory('My', MySchema())
        d = MyDataProxy(data={'a': 1})
        assert d.get('b') is missing
        assert d._data['in_mongo_b'] is missing
        d.set('b', 2)
        assert d.get('b') == 2
        d.delete('b')
        # Can do it two time in a row without error
        d.delete('b')
        assert d._data['in_mongo_b'] is missing

    def test_default(self):
        default_value = ObjectId('507f1f77bcf86cd799439011')
        default_callable = lambda: ObjectId('507f1f77bcf86cd799439012')

        class MySchema(Schema):
            no_default = fields.ObjectIdField()
            with_default = fields.ObjectIdField(default=default_value)
            with_callable_default = fields.ObjectIdField(default=default_callable)

        MyDataProxy = data_proxy_factory('My', MySchema())
        d = MyDataProxy(data={})
        assert d._data['no_default'] is missing
        assert d._data['with_default'] == default_value
        assert d._data['with_callable_default'] == default_callable()
        assert d.get('no_default') is missing
        assert d.get('with_default') == default_value
        assert d.get('with_callable_default') == default_callable()
        assert d.to_mongo() == {
            'with_default': default_value,
            'with_callable_default': default_callable(),
        }
        assert d.dump() == {
            'with_default': str(default_value),
            'with_callable_default': str(default_callable()),
        }

        d.delete('with_default')
        assert d._data['with_default'] == default_value
        assert d.get('with_default') == default_value
        d.delete('with_callable_default')
        assert d._data['with_callable_default'] == default_callable()
        assert d.get('with_callable_default') == default_callable()

    def test_validate(self):

        class MySchema(Schema):
            with_max = fields.IntField(validate=validate.Range(max=99))

        MyDataProxy = data_proxy_factory('My', MySchema())
        d = MyDataProxy(data={})
        with pytest.raises(ValidationError) as exc:
            MyDataProxy(data={'with_max': 100})
        assert exc.value.args[0] == {'with_max': ['Must be less than or equal to 99.']}
        with pytest.raises(ValidationError) as exc:
            d.set('with_max', 100)
        assert exc.value.args[0] == ['Must be less than or equal to 99.']

    def test_required_validate(self):

        @self.instance.register
        class MyEmbedded(EmbeddedDocument):
            required = fields.IntField(required=True)

        class MySchema(Schema):
            # EmbeddedField need instance to retrieve implementation
            required = fields.IntField(required=True)
            embedded = fields.EmbeddedField(MyEmbedded, instance=self.instance)
            listed = fields.ListField(fields.EmbeddedField(MyEmbedded, instance=self.instance))
            dicted = fields.DictField(
                values=fields.EmbeddedField(MyEmbedded, instance=self.instance))

        MyDataProxy = data_proxy_factory('My', MySchema())
        d = MyDataProxy()

        d.load({
            'required': 42,
            'embedded': {'required': 42},
            'listed': [{'required': 42}],
            'dicted': {'a': {'required': 42}}
        })
        d.required_validate()
        # Empty list/dict should not trigger required if embedded field has required fields
        d.load({'embedded': {'required': 42}, 'required': 42})
        d.required_validate()

        d.load({'embedded': {'required': 42}})
        with pytest.raises(ValidationError) as exc:
            d.required_validate()
        assert exc.value.messages == {'required': ['Missing data for required field.']}

        # Missing embedded is valid even though some fields are required in the embedded document
        d.load({'required': 42})
        d.required_validate()
        # Required fields in the embedded document are only checked if the document is not missing
        d.load({'embedded': {}, 'required': 42})
        with pytest.raises(ValidationError) as exc:
            d.required_validate()
        assert exc.value.messages == {'embedded': {'required': ['Missing data for required field.']}}

        d.load({'embedded': {'required': 42}, 'required': 42, 'listed': [{}], 'dicted': {'a': {}}})
        with pytest.raises(ValidationError) as exc:
            d.required_validate()
        assert exc.value.messages == {
            'listed': {0: {'required': ['Missing data for required field.']}},
            'dicted': {'a': {'value': {'required': ['Missing data for required field.']}}},
        }

    def test_unkown_field_in_db(self):
        class MySchema(Schema):
            field = fields.IntField(attribute='mongo_field')

        DataProxy = data_proxy_factory('My', MySchema())
        d = DataProxy()
        d.from_mongo({'mongo_field': 42})
        assert d._data == {'mongo_field': 42}
        with pytest.raises(exceptions.UnknownFieldInDBError):
            d.from_mongo({'mongo_field': 42, 'xxx': 'foo'})

    def test_iterators(self):
        class MySchema(Schema):
            field_a = fields.IntField(attribute='mongo_field_a')
            field_b = fields.IntField(attribute='mongo_field_b')

        DataProxy = data_proxy_factory('My', MySchema())
        d = DataProxy()
        d.from_mongo({'mongo_field_a': 42, 'mongo_field_b': 24})

        assert set(d.keys()) == {'mongo_field_a', 'mongo_field_b'}
        assert set(d.values()) == {42, 24}
        assert set(d.items()) == {('field_a', 42), ('field_b', 24)}

        d.load({'field_a': 100, 'field_b': 200})
        assert set(d.keys()) == {'mongo_field_a', 'mongo_field_b'}
        assert set(d.values()) == {100, 200}
        assert set(d.items()) == {('field_a', 100), ('field_b', 200)}


class TestNonStrictDataProxy(BaseTest):

    def test_build(self):

        class MySchema(Schema):
            pass

        strict_proxy = data_proxy_factory('My', MySchema(), strict=True)
        assert issubclass(strict_proxy, BaseDataProxy)
        non_strict_proxy = data_proxy_factory('My', MySchema(), strict=False)
        assert issubclass(non_strict_proxy, BaseNonStrictDataProxy)

    def test_basic(self):
        class MySchema(Schema):
            field_a = fields.IntField(attribute='mongo_field_a')

        NonStrictDataProxy = data_proxy_factory('My', MySchema(), strict=False)
        with pytest.raises(exceptions.ValidationError) as exc:
            NonStrictDataProxy({'field_a': 42, 'xxx': 'foo'})
        assert exc.value.messages == {'xxx': ['Unknown field.']}
        d = NonStrictDataProxy()
        d.from_mongo({'mongo_field_a': 42, 'xxx': 'foo'})
        assert d._data == {'mongo_field_a': 42}
        assert d._additional_data == {'xxx': 'foo'}



from copy import copy, deepcopy

import pytest

from marshmallow import ValidationError, missing

from umongo.data_proxy import data_proxy_factory
from umongo import Document, EmbeddedDocument, fields, exceptions

from .common import BaseTest


class TestEmbeddedDocument(BaseTest):

    def test_embedded_inheritance(self):
        @self.instance.register
        class MyChildEmbeddedDocument(EmbeddedDocument):
            num = fields.IntField()

        @self.instance.register
        class MyParentEmbeddedDocument(EmbeddedDocument):
            embedded = fields.EmbeddedField(MyChildEmbeddedDocument)

        @self.instance.register
        class MyDoc(Document):
            embedded = fields.EmbeddedField(MyParentEmbeddedDocument)

        document = MyDoc(**{
            "embedded": {"embedded": {"num": 1}}
        })

        assert document is not None

    def test_embedded_document(self):
        @self.instance.register
        class MyEmbeddedDocument(EmbeddedDocument):
            a = fields.IntField(attribute='in_mongo_a')
            b = fields.IntField()

        embedded = MyEmbeddedDocument()
        assert embedded.to_mongo(update=True) is None
        assert not embedded.is_modified()

        @self.instance.register
        class MyDoc(Document):
            embedded = fields.EmbeddedField(MyEmbeddedDocument,
                attribute='in_mongo_embedded', allow_none=True)

        MySchema = MyDoc.Schema

        # Make sure embedded document doesn't have implicit _id field
        assert '_id' not in MyEmbeddedDocument.Schema().fields
        assert 'id' not in MyEmbeddedDocument.Schema().fields

        MyDataProxy = data_proxy_factory('My', MySchema())
        d = MyDataProxy()
        d.from_mongo(data={'in_mongo_embedded': {'in_mongo_a': 1, 'b': 2}})
        assert d.dump() == {'embedded': {'a': 1, 'b': 2}}
        embedded = d.get('embedded')
        assert type(embedded) == MyEmbeddedDocument
        assert embedded.a == 1
        assert embedded.b == 2
        assert embedded.dump() == {'a': 1, 'b': 2}
        assert embedded.to_mongo() == {'in_mongo_a': 1, 'b': 2}
        assert d.to_mongo() == {'in_mongo_embedded': {'in_mongo_a': 1, 'b': 2}}

        d2 = MyDataProxy()
        d2.from_mongo(data={'in_mongo_embedded': {'in_mongo_a': 1, 'b': 2}})
        assert d == d2

        embedded.a = 3
        assert embedded.is_modified()
        assert embedded.to_mongo(update=True) == {'$set': {'in_mongo_a': 3}}
        assert d.to_mongo(update=True) == {'$set': {'in_mongo_embedded': {'in_mongo_a': 3, 'b': 2}}}
        embedded.clear_modified()
        assert embedded.to_mongo(update=True) is None
        assert d.to_mongo(update=True) is None

        del embedded.a
        assert embedded.to_mongo(update=True) == {'$unset': {'in_mongo_a': ''}}
        assert d.to_mongo(update=True) == {'$set': {'in_mongo_embedded': {'b': 2}}}

        d.set('embedded', MyEmbeddedDocument(a=4))
        assert d.get('embedded').to_mongo(update=True) == {'$set': {'in_mongo_a': 4}}
        d.get('embedded').clear_modified()
        assert d.get('embedded').to_mongo(update=True) is None
        assert d.to_mongo(update=True) == {'$set': {'in_mongo_embedded': {'in_mongo_a': 4}}}

        embedded_doc = MyEmbeddedDocument(a=1, b=2)
        assert embedded_doc.a == 1
        assert embedded_doc.b == 2
        assert embedded_doc == {'in_mongo_a': 1, 'b': 2}
        assert embedded_doc == MyEmbeddedDocument(a=1, b=2)
        assert embedded_doc['a'] == 1
        assert embedded_doc['b'] == 2

        embedded_doc.clear_modified()
        embedded_doc.update({'b': 42})
        assert embedded_doc.is_modified()
        assert embedded_doc.a == 1
        assert embedded_doc.b == 42

        with pytest.raises(ValidationError):
            MyEmbeddedDocument(in_mongo_a=1, b=2)

        embedded_doc['a'] = 1
        assert embedded_doc.a == embedded_doc['a'] == 1
        del embedded_doc['a']
        assert embedded_doc.a is embedded_doc['a'] is None

        # Test repr readability
        repr_d = repr(MyEmbeddedDocument(a=1, b=2))
        assert 'tests.test_embedded_document.MyEmbeddedDocument' in repr_d
        assert "'in_mongo_a'" not in repr_d
        assert "'a': 1" in repr_d
        assert "'b': 2" in repr_d

        # Test unknown fields
        with pytest.raises(AttributeError):
            embedded_doc.dummy
        with pytest.raises(AttributeError):
            embedded_doc.dummy = None
        with pytest.raises(AttributeError):
            del embedded_doc.dummy
        with pytest.raises(KeyError):
            embedded_doc['dummy']
        with pytest.raises(KeyError):
            embedded_doc['dummy'] = None
        with pytest.raises(KeyError):
            del embedded_doc['dummy']

        # Test allow_none
        d3 = MyDataProxy({'embedded': None})
        assert d3.to_mongo() == {'in_mongo_embedded': None}
        d3.from_mongo({'in_mongo_embedded': None})
        assert d3.get('embedded') is None

    def test_bad_embedded_document(self):

        @self.instance.register
        class MyEmbeddedDocument(EmbeddedDocument):
            a = fields.IntField()

        @self.instance.register
        class MyDoc(Document):
            e = fields.EmbeddedField(MyEmbeddedDocument)
            l = fields.ListField(fields.EmbeddedField(MyEmbeddedDocument))
            b = fields.IntField(required=True)

        with pytest.raises(ValidationError) as exc:
            MyDoc(l={})
        assert exc.value.args[0] == {'l': ['Not a valid list.']}

        with pytest.raises(ValidationError) as exc:
            MyDoc(l=True)
        assert exc.value.args[0] == {'l': ['Not a valid list.']}

        with pytest.raises(ValidationError) as exc:
            MyDoc(l="string is not a list")
        assert exc.value.args[0] == {'l': ['Not a valid list.']}

        with pytest.raises(ValidationError) as exc:
            MyDoc(l=[42])
        assert exc.value.args[0] == {'l': {0: {'_schema': ['Invalid input type.']}}}

        with pytest.raises(ValidationError) as exc:
            MyDoc(l=[{}, 42])
        assert exc.value.args[0] == {'l': {1: {'_schema': ['Invalid input type.']}}}

        with pytest.raises(ValidationError) as exc:
            MyDoc(b=[{}])
        assert exc.value.args[0] == {'b': ['Not a valid integer.']}

        with pytest.raises(ValidationError) as exc:
            MyDoc(e=[{}])
        assert exc.value.args[0] == {'e': {'_schema': ['Invalid input type.']}}

    def test_inheritance(self):
        @self.instance.register
        class EmbeddedParent(EmbeddedDocument):
            a = fields.IntField(attribute='in_mongo_a_parent')
            b = fields.IntField()

        @self.instance.register
        class EmbeddedChild(EmbeddedParent):
            a = fields.IntField(attribute='in_mongo_a_child')
            c = fields.IntField()

        @self.instance.register
        class GrandChild(EmbeddedChild):
            d = fields.IntField()

        @self.instance.register
        class OtherEmbedded(EmbeddedDocument):
            pass

        @self.instance.register
        class MyDoc(Document):
            parent = fields.EmbeddedField(EmbeddedParent)
            child = fields.EmbeddedField(EmbeddedChild)

        assert EmbeddedParent.opts.offspring == {EmbeddedChild, GrandChild}
        assert EmbeddedChild.opts.offspring == {GrandChild}
        assert GrandChild.opts.offspring == set()
        assert OtherEmbedded.opts.offspring == set()

        parent = EmbeddedParent(a=1)
        child = EmbeddedChild(a=1, b=2, c=3)
        grandchild = GrandChild(d=4)

        assert parent.to_mongo() == {'in_mongo_a_parent': 1}
        assert child.to_mongo() == {'in_mongo_a_child': 1, 'b': 2, 'c': 3, '_cls': 'EmbeddedChild'}
        assert grandchild.to_mongo() == {'d': 4, '_cls': 'GrandChild'}

        with pytest.raises(ValidationError) as exc:
            MyDoc(parent=OtherEmbedded())
        assert exc.value.args[0] == {'parent': {'_schema': ['Invalid input type.']}}
        with pytest.raises(ValidationError):
            MyDoc(child=parent)
        doc = MyDoc(parent=child, child=child)
        assert doc.child == doc.parent

        doc = MyDoc(child={'a': 1, 'cls': 'GrandChild'},
                    parent={'cls': 'EmbeddedChild', 'a': 1})
        assert doc.child.to_mongo() == {'in_mongo_a_child': 1, '_cls': 'GrandChild'}
        assert doc.parent.to_mongo() == {'in_mongo_a_child': 1, '_cls': 'EmbeddedChild'}

        with pytest.raises(ValidationError) as exc:
            MyDoc(child={'a': 1, '_cls': 'GrandChild'})
        assert exc.value.messages == {'child': {'_cls': ['Unknown field.']}}

        # Try to build a non-child document
        with pytest.raises(ValidationError) as exc:
            MyDoc(child={'cls': 'OtherEmbedded'})
        assert exc.value.messages == {'child': ['Unknown document `OtherEmbedded`.']}

        # Test embedded child deserialization from mongo
        child = EmbeddedChild(c=69)
        doc = MyDoc(parent=child)
        mongo_data = doc.to_mongo()
        doc2 = MyDoc.build_from_mongo(mongo_data)
        assert isinstance(doc2.parent, EmbeddedChild)
        assert doc._data == doc2._data

        # Test grandchild can be passed as parent
        doc = MyDoc(parent={'cls': 'GrandChild', 'd': 2})
        assert doc.parent.to_mongo() == {'d': 2, '_cls': 'GrandChild'}

    def test_abstract_inheritance(self):
        class AlienClass:
            pass

        @self.instance.register
        class AbstractParent(EmbeddedDocument):
            a = fields.IntField(attribute='in_mongo_a_parent')
            b = fields.IntField()

            class Meta:
                abstract = True

        @self.instance.register
        class AbstractChild(AbstractParent):
            a = fields.IntField(attribute='in_mongo_a_child')
            c = fields.IntField()

            class Meta:
                abstract = True

        @self.instance.register
        class ConcreteChild(AbstractParent, AlienClass):
            c = fields.IntField()

            class Meta:
                allow_inheritance = True

        @self.instance.register
        class ConcreteGrandChild(AbstractChild):
            d = fields.IntField()

        @self.instance.register
        class ConcreteConcreteGrandChild(ConcreteChild):
            d = fields.IntField()

        @self.instance.register
        class OtherEmbedded(EmbeddedDocument):
            pass

        with pytest.raises(exceptions.AbstractDocumentError) as exc:
            AbstractParent()
        assert exc.value.args[0] == "Cannot instantiate an abstract EmbeddedDocument"

        # test
        cc = ConcreteChild(a=1, b=2, c=3)
        cgc = ConcreteGrandChild(a=1, b=2, c=3, d=4)
        ccgc = ConcreteConcreteGrandChild(a=1, b=2, c=3, d=4)

        assert cc.to_mongo() == {'in_mongo_a_parent': 1, 'b': 2, 'c': 3, '_cls': 'ConcreteChild'}
        assert cgc.to_mongo() == {'in_mongo_a_child': 1, 'b': 2, 'c': 3, 'd': 4, '_cls': 'ConcreteGrandChild'}
        assert ccgc.to_mongo() == {'in_mongo_a_parent': 1, 'b': 2, 'c': 3, 'd': 4, '_cls': 'ConcreteConcreteGrandChild'}

    def test_bad_inheritance(self):
        with pytest.raises(exceptions.DocumentDefinitionError) as exc:
            @self.instance.register
            class BadAbstract(EmbeddedDocument):
                class Meta:
                    allow_inheritance = False
                    abstract = True
        assert exc.value.args[0] == "Abstract embedded document cannot disable inheritance"

        @self.instance.register
        class NotParent(EmbeddedDocument):
            class Meta:
                allow_inheritance = False

        with pytest.raises(exceptions.DocumentDefinitionError) as exc:
            @self.instance.register
            class ImpossibleChild1(NotParent):
                pass
        assert exc.value.args[0] == ("EmbeddedDocument"
            " <Implementation class 'tests.test_embedded_document.NotParent'>"
            " doesn't allow inheritance")

        @self.instance.register
        class NotAbstractParent(EmbeddedDocument):
            class Meta:
                allow_inheritance = True

        # Unlike Document, EmbeddedDocument should allow inheritance by default
        assert NotAbstractParent.opts.allow_inheritance

        with pytest.raises(exceptions.DocumentDefinitionError) as exc:
            @self.instance.register
            class ImpossibleChild2(NotAbstractParent):
                class Meta:
                    abstract = True
        assert exc.value.args[0] == "Abstract embedded document should have all it parents abstract"

    def test_property(self):
        @self.instance.register
        class MyEmbeddedDoc(EmbeddedDocument):
            _prop = fields.FloatField()

            @property
            def prop(self):
                return self._prop

            @prop.setter
            def prop(self, value):
                self._prop = value

            @prop.deleter
            def prop(self):
                del self._prop

        emb = MyEmbeddedDoc()
        emb.prop = 42
        assert emb.prop == 42
        del emb.prop
        assert emb.prop is None

    def test_equality(self):
        @self.instance.register
        class MyChildEmbeddedDocument(EmbeddedDocument):
            num = fields.IntField()

        @self.instance.register
        class MyParentEmbeddedDocument(EmbeddedDocument):
            embedded = fields.EmbeddedField(MyChildEmbeddedDocument)

        emb_1 = MyParentEmbeddedDocument(embedded={'num': 1})
        emb_2 = MyParentEmbeddedDocument(embedded={'num': 1})
        emb_3 = MyParentEmbeddedDocument(embedded={})
        emb_4 = MyParentEmbeddedDocument()

        assert emb_1 == emb_2
        assert emb_1 != emb_3
        assert emb_1 != emb_4
        assert emb_1 != None  # noqa: E711 (None comparison)
        assert emb_1 != missing
        assert None != emb_1  # noqa: E711 (None comparison)
        assert missing != emb_1

    def test_strict_embedded_document(self):
        @self.instance.register
        class StrictEmbeddedDoc(EmbeddedDocument):
            a = fields.IntField()

        @self.instance.register
        class NonStrictEmbeddedDoc(EmbeddedDocument):
            a = fields.IntField()

            class Meta:
                strict = False

        data_with_bonus = {'a': 42, 'b': 'foo'}
        with pytest.raises(exceptions.UnknownFieldInDBError):
            StrictEmbeddedDoc.build_from_mongo(data_with_bonus)

        non_strict_doc = NonStrictEmbeddedDoc.build_from_mongo(data_with_bonus)
        assert non_strict_doc.to_mongo() == data_with_bonus
        non_strict_doc.dump() == {'a': 42}

        with pytest.raises(exceptions.ValidationError) as exc:
            NonStrictEmbeddedDoc(a=42, b='foo')
        assert exc.value.messages == {'b': ['Unknown field.']}

    def test_deepcopy(self):

        @self.instance.register
        class Child(EmbeddedDocument):
            name = fields.StrField()

        @self.instance.register
        class Parent(EmbeddedDocument):
            name = fields.StrField()
            child = fields.EmbeddedField(Child)

        john = Parent(name='John Doe', child={'name': 'John Doe Jr.'})
        jane = copy(john)
        assert jane.name == john.name
        assert jane.child is john.child
        jane = deepcopy(john)
        assert jane.name == john.name
        assert jane.child == john.child
        assert jane.child is not john.child



from marshmallow import ValidationError, missing

from .abstract import BaseDataObject
from .exceptions import UnknownFieldInDBError
from .i18n import gettext as _


__all__ = ('data_proxy_factory', 'missing')


class BaseDataProxy:

    __slots__ = ('_data', '_modified_data')
    schema = None
    _fields = None
    _fields_from_mongo_key = None

    def __init__(self, data=None):
        # Inside data proxy, data are stored in mongo world representation
        self._modified_data = set()
        self._data = {}
        self.load(data or {})

    def to_mongo(self, update=False):
        if update:
            return self._to_mongo_update()
        return self._to_mongo()

    def _to_mongo(self):
        mongo_data = {}
        for key, val in self._data.items():
            field = self._fields_from_mongo_key[key]
            val = field.serialize_to_mongo(val)
            if val is not missing:
                mongo_data[key] = val
        return mongo_data

    def _to_mongo_update(self):
        mongo_data = {}
        set_data = {}
        unset_data = []
        for name in self.get_modified_fields():
            field = self._fields[name]
            name = field.attribute or name
            val = field.serialize_to_mongo(self._data[name])
            if val is missing:
                unset_data.append(name)
            else:
                set_data[name] = val
        if set_data:
            mongo_data['$set'] = set_data
        if unset_data:
            mongo_data['$unset'] = {k: "" for k in unset_data}
        return mongo_data or None

    def from_mongo(self, data):
        self._data = {}
        for key, val in data.items():
            try:
                field = self._fields_from_mongo_key[key]
            except KeyError:
                raise UnknownFieldInDBError(_(
                    '{cls}: unknown "{key}" field found in DB.'
                    .format(key=key, cls=self.__class__.__name__)
                ))
            self._data[key] = field.deserialize_from_mongo(val)
        self.clear_modified()
        self._add_missing_fields()

    def dump(self):
        return self.schema.dump(self._data)

    def _mark_as_modified(self, key):
        self._modified_data.add(key)

    def update(self, data):
        # Always use marshmallow partial load to skip required checks
        loaded_data = self.schema.load(data, partial=True)
        self._data.update(loaded_data)
        for key in loaded_data:
            self._mark_as_modified(key)

    def load(self, data):
        # Always use marshmallow partial load to skip required checks
        loaded_data = self.schema.load(data, partial=True)
        self._data = loaded_data
        # Map the modified fields list on the the loaded data
        self.clear_modified()
        for key in loaded_data:
            self._mark_as_modified(key)
        # TODO: mark added missing fields as modified?
        self._add_missing_fields()

    def _get_field(self, name, to_raise):
        if name not in self._fields:
            raise to_raise(name)
        field = self._fields[name]
        name = field.attribute or name
        return name, field

    def get(self, name, to_raise=KeyError):
        name, _ = self._get_field(name, to_raise)
        return self._data[name]

    def set(self, name, value, to_raise=KeyError):
        name, field = self._get_field(name, to_raise)
        if value is None and not getattr(field, 'allow_none', False):
            raise ValidationError(field.error_messages['null'])
        if value is not None:
            value = field._deserialize(value, name, None)
            field._validate(value)
        self._data[name] = value
        self._mark_as_modified(name)

    def delete(self, name, to_raise=KeyError):
        name, field = self._get_field(name, to_raise)
        default = field.default
        self._data[name] = default() if callable(default) else default
        self._mark_as_modified(name)

    def __repr__(self):
        # Display data in oo world format
        return "<%s(%s)>" % (self.__class__.__name__, dict(self.items()))

    def __eq__(self, other):
        if isinstance(other, dict):
            return self._data == other
        if hasattr(other, '_data'):
            return self._data == other._data
        return NotImplemented

    def get_modified_fields(self):
        modified = set()
        for name, field in self._fields.items():
            value_name = field.attribute or name
            value = self._data[value_name]
            if value_name in self._modified_data or (
                    isinstance(value, BaseDataObject) and value.is_modified()):
                modified.add(name)
        return modified

    def clear_modified(self):
        self._modified_data.clear()
        for val in self._data.values():
            if isinstance(val, BaseDataObject):
                val.clear_modified()

    def is_modified(self):
        return (
            bool(self._modified_data) or
            any(isinstance(v, BaseDataObject) and v.is_modified()
                for v in self._data.values())
        )

    def _add_missing_fields(self):
        # TODO: we should be able to do that by configuring marshmallow...
        for name, field in self._fields.items():
            mongo_name = field.attribute or name
            if mongo_name not in self._data:
                if callable(field.missing):
                    self._data[mongo_name] = field.missing()
                else:
                    self._data[mongo_name] = field.missing

    def required_validate(self):
        errors = {}
        for name, field in self.schema.fields.items():
            value = self._data[field.attribute or name]
            if field.required and value is missing:
                errors[name] = [_("Missing data for required field.")]
            elif hasattr(field, '_required_validate'):
                try:
                    field._required_validate(value)
                except ValidationError as exc:
                    errors[name] = exc.messages
        if errors:
            raise ValidationError(errors)

    # Standards iterators providing oo and mongo worlds views

    def items(self):
        return (
            (key, self._data[field.attribute or key]) for key, field in self._fields.items()
        )

    def keys(self):
        return (field.attribute or key for key, field in self._fields.items())

    def values(self):
        return self._data.values()


class BaseNonStrictDataProxy(BaseDataProxy):
    """
    This data proxy will accept unknown data comming from mongo and will
    return them along with other data when ask.
    """

    __slots__ = ('_additional_data', )

    def __init__(self, data=None):
        self._additional_data = {}
        super().__init__(data=data)

    def _to_mongo(self):
        mongo_data = super()._to_mongo()
        mongo_data.update(self._additional_data)
        return mongo_data

    def from_mongo(self, data):
        self._data = {}
        for key, val in data.items():
            try:
                field = self._fields_from_mongo_key[key]
            except KeyError:
                self._additional_data[key] = val
            else:
                self._data[key] = field.deserialize_from_mongo(val)
        self.clear_modified()
        self._add_missing_fields()


def data_proxy_factory(basename, schema, strict=True):
    """
    Generate a DataProxy from the given schema.

    This way all generic informations (like schema and fields lookups)
    are kept inside the  DataProxy class and it instances are just flyweights.
    """

    cls_name = "%sDataProxy" % basename

    nmspc = {
        '__slots__': (),
        'schema': schema,
        '_fields': schema.fields,
        '_fields_from_mongo_key': {v.attribute or k: v for k, v in schema.fields.items()}
    }

    data_proxy_cls = type(cls_name, (BaseDataProxy if strict else BaseNonStrictDataProxy, ), nmspc)
    return data_proxy_cls



from inspect import iscoroutine
import asyncio

from motor.motor_asyncio import AsyncIOMotorDatabase, AsyncIOMotorCursor
from motor import version_tuple as MOTOR_VERSION
from pymongo.errors import DuplicateKeyError

from ..builder import BaseBuilder
from ..document import DocumentImplementation
from ..data_proxy import missing
from ..data_objects import Reference
from ..exceptions import (
    NotCreatedError, UpdateError, DeleteError, ValidationError, NoneReferenceError
)
from ..fields import ReferenceField, ListField, EmbeddedField
from ..query_mapper import map_query

from .tools import cook_find_filter


class WrappedCursor(AsyncIOMotorCursor):

    __slots__ = ('raw_cursor', 'document_cls')

    def __init__(self, document_cls, cursor):
        # Such a cunning plan my lord !
        # We inherit from Cursor but don't call it __init__ because
        # we act as a proxy to the underlying raw_cursor
        WrappedCursor.raw_cursor.__set__(self, cursor)
        WrappedCursor.document_cls.__set__(self, document_cls)

    def __getattr__(self, name):
        return getattr(self.raw_cursor, name)

    def __setattr__(self, name, value):
        return setattr(self.raw_cursor, name, value)

    def clone(self):
        return WrappedCursor(self.document_cls, self.raw_cursor.clone())

    def next_object(self):
        raw = self.raw_cursor.next_object()
        return self.document_cls.build_from_mongo(raw, use_cls=True)

    def each(self, callback):
        def wrapped_callback(result, error):
            if not error and result is not None:
                result = self.document_cls.build_from_mongo(result, use_cls=True)
            return callback(result, error)
        return self.raw_cursor.each(wrapped_callback)

    def to_list(self, length, callback=None):
        kwargs = {"callback": callback} if callback else {}
        raw_future = self.raw_cursor.to_list(length, **kwargs)
        cooked_future = asyncio.Future()
        builder = self.document_cls.build_from_mongo

        def on_raw_done(fut):
            cooked_future.set_result([builder(e, use_cls=True) for e in fut.result()])

        raw_future.add_done_callback(on_raw_done)
        return cooked_future


class MotorAsyncIODocument(DocumentImplementation):

    __slots__ = ()

    opts = DocumentImplementation.opts

    # Cook hooks into coroutines in order to allow them to return
    # either Future or regular return value.

    async def __coroutined_pre_insert(self):
        ret = self.pre_insert()
        if iscoroutine(ret):
            ret = await ret
        return ret

    async def __coroutined_pre_update(self):
        ret = self.pre_update()
        if iscoroutine(ret):
            ret = await ret
        return ret

    async def __coroutined_pre_delete(self):
        ret = self.pre_delete()
        if iscoroutine(ret):
            ret = await ret
        return ret

    async def __coroutined_post_insert(self, ret):
        ret = self.post_insert(ret)
        if iscoroutine(ret):
            ret = await ret
        return ret

    async def __coroutined_post_update(self, ret):
        ret = self.post_update(ret)
        if iscoroutine(ret):
            ret = await ret
        return ret

    async def __coroutined_post_delete(self, ret):
        ret = self.post_delete(ret)
        if iscoroutine(ret):
            ret = await ret
        return ret

    async def reload(self):
        """
        Retrieve and replace document's data by the ones in database.

        Raises :class:`umongo.exceptions.NotCreatedError` if the document
        doesn't exist in database.
        """
        if not self.is_created:
            raise NotCreatedError("Document doesn't exists in database")
        ret = await self.collection.find_one(self.pk)
        if ret is None:
            raise NotCreatedError("Document doesn't exists in database")
        self._data = self.DataProxy()
        self._data.from_mongo(ret)

    async def commit(self, io_validate_all=False, conditions=None):
        """
        Commit the document in database.
        If the document doesn't already exist it will be inserted, otherwise
        it will be updated.

        :param io_validate_all:
        :param conditions: only perform commit if matching record in db
            satisfies condition(s) (e.g. version number).
            Raises :class:`umongo.exceptions.UpdateError` if the
            conditions are not satisfied.
        :return: Update result dict returned by underlaying driver or
            ObjectId of the inserted document.
        """
        try:
            if self.is_created:
                if self.is_modified():
                    query = conditions or {}
                    query['_id'] = self.pk
                    # pre_update can provide additional query filter and/or
                    # modify the fields' values
                    additional_filter = await self.__coroutined_pre_update()
                    if additional_filter:
                        query.update(map_query(additional_filter, self.schema.fields))
                    self.required_validate()
                    await self.io_validate(validate_all=io_validate_all)
                    payload = self._data.to_mongo(update=True)
                    ret = await self.collection.update_one(query, payload)
                    if ret.matched_count != 1:
                        raise UpdateError(ret)
                    await self.__coroutined_post_update(ret)
                else:
                    ret = None
            elif conditions:
                raise RuntimeError('Document must already exist in database to use `conditions`.')
            else:
                await self.__coroutined_pre_insert()
                self.required_validate()
                await self.io_validate(validate_all=io_validate_all)
                payload = self._data.to_mongo(update=False)
                ret = await self.collection.insert_one(payload)
                # TODO: check ret ?
                self._data.set(self.pk_field, ret.inserted_id)
                self.is_created = True
                await self.__coroutined_post_insert(ret)
        except DuplicateKeyError as exc:
            # Need to dig into error message to find faulting index
            errmsg = exc.details['errmsg']
            for index in self.opts.indexes:
                if ('.$%s' % index.document['name'] in errmsg or
                        ' %s ' % index.document['name'] in errmsg):
                    keys = index.document['key'].keys()
                    if len(keys) == 1:
                        key = tuple(keys)[0]
                        msg = self.schema.fields[key].error_messages['unique']
                        raise ValidationError({key: msg})
                    fields = self.schema.fields
                    # Compound index (sort value to make testing easier)
                    keys = sorted(keys)
                    raise ValidationError({
                        k: fields[k].error_messages['unique_compound'].format(fields=keys)
                        for k in keys
                    })
            # Unknown index, cannot wrap the error so just reraise it
            raise
        self._data.clear_modified()
        return ret

    async def delete(self, conditions=None):
        """
        Alias of :meth:`remove` to enforce default api.
        """
        return await self.remove(conditions=conditions)

    async def remove(self, conditions=None):
        """
        Remove the document from database.

        :param conditions: Only perform delete if matching record in db
            satisfies condition(s) (e.g. version number).
            Raises :class:`umongo.exceptions.DeleteError` if the
            conditions are not satisfied.
        Raises :class:`umongo.exceptions.NotCreatedError` if the document
        is not created (i.e. ``doc.is_created`` is False)
        Raises :class:`umongo.exceptions.DeleteError` if the document
        doesn't exist in database.

        :return: Delete result dict returned by underlaying driver.
        """
        if not self.is_created:
            raise NotCreatedError("Document doesn't exists in database")
        query = conditions or {}
        query['_id'] = self.pk
        # pre_delete can provide additional query filter
        additional_filter = await self.__coroutined_pre_delete()
        if additional_filter:
            query.update(map_query(additional_filter, self.schema.fields))
        ret = await self.collection.delete_one(query)
        if ret.deleted_count != 1:
            raise DeleteError(ret)
        self.is_created = False
        await self.__coroutined_post_delete(ret)
        return ret

    async def io_validate(self, validate_all=False):
        """
        Run the io_validators of the document's fields.

        :param validate_all: If False only run the io_validators of the
            fields that have been modified.
        """
        if validate_all:
            return await _io_validate_data_proxy(self.schema, self._data)
        return await _io_validate_data_proxy(
            self.schema, self._data, partial=self._data.get_modified_fields())

    @classmethod
    async def find_one(cls, filter=None, *args, **kwargs):
        """
        Find a single document in database.
        """
        filter = cook_find_filter(cls, filter)
        ret = await cls.collection.find_one(filter, *args, **kwargs)
        if ret is not None:
            ret = cls.build_from_mongo(ret, use_cls=True)
        return ret

    @classmethod
    def find(cls, filter=None, *args, **kwargs):
        """
        Find a list document in database.

        Returns a cursor that provide Documents.
        """
        filter = cook_find_filter(cls, filter)
        return WrappedCursor(cls, cls.collection.find(filter, *args, **kwargs))

    @classmethod
    async def count_documents(cls, filter=None, *, with_limit_and_skip=False, **kwargs):
        """
        Return a count of the documents in a collection.
        """
        filter = cook_find_filter(cls, filter or {})
        if MOTOR_VERSION < (2, 0):
            count = await cls.find(filter, **kwargs).count(with_limit_and_skip)
            return count
        count = await cls.collection.count_documents(filter, **kwargs)
        return count

    @classmethod
    async def ensure_indexes(cls):
        """
        Check&create if needed the Document's indexes in database
        """
        for index in cls.opts.indexes:
            kwargs = index.document.copy()
            keys = kwargs.pop('key').items()
            await cls.collection.create_index(keys, **kwargs)


# Run multiple validators and collect all errors in one
async def _run_validators(validators, field, value):
    errors = []
    tasks = [validator(field, value) for validator in validators]
    results = await asyncio.gather(*tasks, return_exceptions=True)
    for i, res in enumerate(results):
        if isinstance(res, ValidationError):
            errors.extend(res.messages)
        elif res:
            raise res
    if errors:
        raise ValidationError(errors)


async def _io_validate_data_proxy(schema, data_proxy, partial=None):
    errors = {}
    tasks = []
    tasks_field_name = []
    for name, field in schema.fields.items():
        if partial and name not in partial:
            continue
        data_name = field.attribute or name
        value = data_proxy._data[data_name]
        if value is missing:
            continue
        try:
            if field.io_validate_recursive:
                await field.io_validate_recursive(field, value)
            if field.io_validate:
                tasks.append(_run_validators(field.io_validate, field, value))
                tasks_field_name.append(name)
        except ValidationError as exc:
            errors[name] = exc.messages
    results = await asyncio.gather(*tasks, return_exceptions=True)
    for i, res in enumerate(results):
        if isinstance(res, ValidationError):
            errors[tasks_field_name[i]] = res.messages
        elif res:
            raise res
    if errors:
        raise ValidationError(errors)


async def _reference_io_validate(field, value):
    await value.fetch(no_data=True)


async def _list_io_validate(field, value):
    validators = field.inner.io_validate
    if not validators or not value:
        return
    tasks = [_run_validators(validators, field.inner, e) for e in value]
    results = await asyncio.gather(*tasks, return_exceptions=True)
    errors = {}
    for i, res in enumerate(results):
        if isinstance(res, ValidationError):
            errors[i] = res.messages
        elif res:
            raise res
    if errors:
        raise ValidationError(errors)


async def _embedded_document_io_validate(field, value):
    await _io_validate_data_proxy(value.schema, value._data)


class MotorAsyncIOReference(Reference):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._document = None

    async def fetch(self, no_data=False, force_reload=False):
        if not self._document or force_reload:
            if self.pk is None:
                raise NoneReferenceError('Cannot retrieve a None Reference')
            self._document = await self.document_cls.find_one(self.pk)
            if not self._document:
                raise ValidationError(self.error_messages['not_found'].format(
                    document=self.document_cls.__name__))
        return self._document


class MotorAsyncIOBuilder(BaseBuilder):

    BASE_DOCUMENT_CLS = MotorAsyncIODocument

    @staticmethod
    def is_compatible_with(db):
        return isinstance(db, AsyncIOMotorDatabase)

    def _patch_field(self, field):
        super()._patch_field(field)

        validators = field.io_validate
        if not validators:
            field.io_validate = []
        else:
            if hasattr(validators, '__iter__'):
                validators = list(validators)
            else:
                validators = [validators]
            field.io_validate = [
                v if asyncio.iscoroutinefunction(v) else asyncio.coroutine(v)
                for v in validators
            ]
        if isinstance(field, ListField):
            field.io_validate_recursive = _list_io_validate
        if isinstance(field, ReferenceField):
            field.io_validate.append(_reference_io_validate)
            field.reference_cls = MotorAsyncIOReference
        if isinstance(field, EmbeddedField):
            field.io_validate_recursive = _embedded_document_io_validate



from bson import DBRef

from .abstract import BaseDataObject, I18nErrorDict
from .i18n import N_


__all__ = ('List', 'Dict', 'Reference')


class List(BaseDataObject, list):

    __slots__ = ('inner_field', '_modified')

    def __init__(self, inner_field, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._modified = False
        self.inner_field = inner_field

    def __setitem__(self, key, obj):
        obj = self.inner_field.deserialize(obj)
        super().__setitem__(key, obj)
        self.set_modified()

    def __delitem__(self, key):
        super().__delitem__(key)
        self.set_modified()

    def append(self, obj):
        obj = self.inner_field.deserialize(obj)
        ret = super().append(obj)
        self.set_modified()
        return ret

    def pop(self, *args, **kwargs):
        ret = super().pop(*args, **kwargs)
        self.set_modified()
        return ret

    def clear(self, *args, **kwargs):
        ret = super().clear(*args, **kwargs)
        self.set_modified()
        return ret

    def remove(self, *args, **kwargs):
        ret = super().remove(*args, **kwargs)
        self.set_modified()
        return ret

    def reverse(self, *args, **kwargs):
        ret = super().reverse(*args, **kwargs)
        self.set_modified()
        return ret

    def sort(self, *args, **kwargs):
        ret = super().sort(*args, **kwargs)
        self.set_modified()
        return ret

    def extend(self, iterable):
        iterable = [self.inner_field.deserialize(obj) for obj in iterable]
        ret = super().extend(iterable)
        self.set_modified()
        return ret

    def __repr__(self):
        return '<object %s.%s(%s)>' % (
            self.__module__, self.__class__.__name__, list(self))

    def is_modified(self):
        if self._modified:
            return True
        if self and isinstance(self[0], BaseDataObject):
            return any(obj.is_modified() for obj in self)
        return False

    def set_modified(self):
        self._modified = True

    def clear_modified(self):
        self._modified = False
        if self and isinstance(self[0], BaseDataObject):
            for obj in self:
                obj.clear_modified()


class Dict(BaseDataObject, dict):

    __slots__ = ('key_field', 'value_field', '_modified')

    def __init__(self, key_field, value_field, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._modified = False
        self.key_field = key_field
        self.value_field = value_field

    def __setitem__(self, key, obj):
        key = self.key_field.deserialize(key) if self.key_field else key
        obj = self.value_field.deserialize(obj) if self.value_field else obj
        super().__setitem__(key, obj)
        self.set_modified()

    def __delitem__(self, key):
        super().__delitem__(key)
        self.set_modified()

    def pop(self, *args, **kwargs):
        ret = super().pop(*args, **kwargs)
        self.set_modified()
        return ret

    def popitem(self, *args, **kwargs):
        ret = super().popitem(*args, **kwargs)
        self.set_modified()
        return ret

    def setdefault(self, key, obj=None):
        key = self.key_field.deserialize(key) if self.key_field else key
        obj = self.value_field.deserialize(obj) if self.value_field else obj
        ret = super().setdefault(key, obj)
        self.set_modified()
        return ret

    def update(self, other):
        new = {
            self.key_field.deserialize(k) if self.key_field else k:
            self.value_field.deserialize(v) if self.value_field else v
            for k, v in other.items()
        }
        super().update(new)
        self.set_modified()

    def __repr__(self):
        return '<object %s.%s(%s)>' % (
            self.__module__, self.__class__.__name__, dict(self))

    def is_modified(self):
        if self._modified:
            return True
        if self and any(isinstance(v, BaseDataObject) for v in self.values()):
            return any(obj.is_modified() for obj in self.values())
        return False

    def set_modified(self):
        self._modified = True

    def clear_modified(self):
        self._modified = False
        if self and any(isinstance(v, BaseDataObject) for v in self.values()):
            for obj in self.values():
                obj.clear_modified()


class Reference:

    error_messages = I18nErrorDict(not_found=N_('Reference not found for document {document}.'))

    def __init__(self, document_cls, pk):
        self.document_cls = document_cls
        self.pk = pk
        self._document = None

    def fetch(self, no_data=False, force_reload=False):
        """
        Retrieve from the database the referenced document

        :param no_data: if True, the caller is only interested in whether or
            not the document is present in database. This means the
            implementation may not retrieve document's data to save bandwidth.
        :param force_reload: if True, ignore any cached data and reload referenced
            document from database.
        """
        raise NotImplementedError
    # TODO replace no_data by `exists` function

    def __repr__(self):
        return '<object %s.%s(document=%s, pk=%r)>' % (
            self.__module__, self.__class__.__name__, self.document_cls.__name__, self.pk)

    def __eq__(self, other):
        if isinstance(other, self.document_cls):
            return other.pk == self.pk
        if isinstance(other, Reference):
            return self.pk == other.pk and self.document_cls == other.document_cls
        if isinstance(other, DBRef):
            return self.pk == other.id and self.document_cls.collection.name == other.collection
        return NotImplemented



from marshmallow import validate as ma_validate

from .abstract import BaseValidator


__all__ = (
    'URL',
    'Email',
    'Range',
    'Length',
    'Equal',
    'Regexp',
    'Predicate',
    'NoneOf',
    'OneOf',
    'ContainsOnly'
)


class URL(BaseValidator, ma_validate.URL):
    pass


class Email(BaseValidator, ma_validate.Email):
    pass


class Range(BaseValidator, ma_validate.Range):
    pass


class Length(BaseValidator, ma_validate.Length):
    pass


class Equal(BaseValidator, ma_validate.Equal):
    pass


class Regexp(BaseValidator, ma_validate.Regexp):
    pass


class Predicate(BaseValidator, ma_validate.Predicate):
    pass


class NoneOf(BaseValidator, ma_validate.NoneOf):
    pass


class OneOf(BaseValidator, ma_validate.OneOf):
    pass


class ContainsOnly(BaseValidator, ma_validate.ContainsOnly):
    pass



import collections
import datetime as dt

from bson import DBRef, ObjectId, Decimal128
from marshmallow import ValidationError, missing
from marshmallow import fields as ma_fields

# from .registerer import retrieve_document
from .document import DocumentImplementation
from .exceptions import NotRegisteredDocumentError
from .template import get_template
from .data_objects import Reference, List, Dict
from . import marshmallow_bonus as ma_bonus_fields
from .abstract import BaseField
from .i18n import gettext as _


__all__ = (
    # 'RawField',
    # 'MappingField',
    # 'TupleField',
    'StringField',
    'UUIDField',
    'NumberField',
    'IntegerField',
    'DecimalField',
    'BooleanField',
    'FloatField',
    'DateTimeField',
    'NaiveDateTimeField',
    'AwareDateTimeField',
    # 'TimeField',
    'DateField',
    # 'TimeDeltaField',
    'UrlField',
    'URLField',
    'EmailField',
    'StrField',
    'BoolField',
    'IntField',
    'DictField',
    'ListField',
    'ConstantField',
    # 'PluckField'
    'ObjectIdField',
    'ReferenceField',
    'GenericReferenceField',
    'EmbeddedField'
)


# Republish supported marshmallow fields


# class RawField(BaseField, ma_fields.Raw):
#     pass


class StringField(BaseField, ma_fields.String):
    pass


class UUIDField(BaseField, ma_fields.UUID):
    pass


class NumberField(BaseField, ma_fields.Number):
    pass


class IntegerField(BaseField, ma_fields.Integer):
    pass


class DecimalField(BaseField, ma_fields.Decimal):
    """
    .. warning::

        This field is only supported on MongoDB 3.4+.
        Using it on older versions will result in uncaught errors.
    """
    def _serialize_to_mongo(self, obj):
        return Decimal128(obj)

    def _deserialize_from_mongo(self, value):
        return value.to_decimal()


class BooleanField(BaseField, ma_fields.Boolean):
    pass


class FloatField(BaseField, ma_fields.Float):
    pass


def _round_to_millisecond(datetime):
    """Round a datetime to millisecond precision

    MongoDB stores datetimes with a millisecond precision.
    For consistency, use the same precision in the object representation.
    """
    microseconds = round(datetime.microsecond, -3)
    if microseconds == 1000000:
        return datetime.replace(microsecond=0) + dt.timedelta(seconds=1)
    return datetime.replace(microsecond=microseconds)


class DateTimeField(BaseField, ma_fields.DateTime):

    def _deserialize(self, value, attr, data, **kwargs):
        if isinstance(value, dt.datetime):
            ret = value
        else:
            ret = super()._deserialize(value, attr, data, **kwargs)
        return _round_to_millisecond(ret)


class NaiveDateTimeField(BaseField, ma_fields.NaiveDateTime):

    def _deserialize(self, value, attr, data, **kwargs):
        if isinstance(value, dt.datetime):
            ret = value
        else:
            ret = super()._deserialize(value, attr, data, **kwargs)
        return _round_to_millisecond(ret)


class AwareDateTimeField(BaseField, ma_fields.AwareDateTime):

    def _deserialize(self, value, attr, data, **kwargs):
        if isinstance(value, dt.datetime):
            ret = value
        else:
            ret = super()._deserialize(value, attr, data, **kwargs)
        return _round_to_millisecond(ret)

    def _deserialize_from_mongo(self, value):
        value = value.replace(tzinfo=dt.timezone.utc)
        if self.default_timezone is not None:
            value = value.astimezone(self.default_timezone)
        return value


# class TimeField(BaseField, ma_fields.Time):
#     pass


class DateField(BaseField, ma_fields.Date):
    """This field converts a date to a datetime to store it as a BSON Date"""

    def _deserialize(self, value, attr, data, **kwargs):
        if isinstance(value, dt.date):
            return value
        return super()._deserialize(value, attr, data)

    def _serialize_to_mongo(self, obj):
        return dt.datetime(obj.year, obj.month, obj.day)

    def _deserialize_from_mongo(self, value):
        return value.date()


# class TimeDeltaField(BaseField, ma_fields.TimeDelta):
#     pass


class UrlField(BaseField, ma_fields.Url):
    pass


class EmailField(BaseField, ma_fields.Email):
    pass


class ConstantField(BaseField, ma_fields.Constant):
    pass


class DictField(BaseField, ma_fields.Dict):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        def cast_value_or_callable(key_field, value_field, value):
            if value is missing:
                return missing
            if callable(value):
                return lambda: Dict(key_field, value_field, value())
            return Dict(key_field, value_field, value)

        self.default = cast_value_or_callable(self.key_field, self.value_field, self.default)
        self.missing = cast_value_or_callable(self.key_field, self.value_field, self.missing)

    def _deserialize(self, value, attr, data, **kwargs):
        value = super()._deserialize(value, attr, data, **kwargs)
        return Dict(self.key_field, self.value_field, value)

    def _serialize_to_mongo(self, obj):
        if obj is None:
            return missing
        return {
            self.key_field.serialize_to_mongo(k) if self.key_field else k:
            self.value_field.serialize_to_mongo(v) if self.value_field else v
            for k, v in obj.items()
        }

    def _deserialize_from_mongo(self, value):
        if value:
            return Dict(
                self.key_field,
                self.value_field,
                {
                    self.key_field.deserialize_from_mongo(k) if self.key_field else k:
                    self.value_field.deserialize_from_mongo(v) if self.value_field else v
                    for k, v in value.items()
                }
            )
        return Dict(self.key_field, self.value_field)

    def as_marshmallow_field(self, mongo_world=False, **kwargs):
        # Overwrite default `as_marshmallow_field` to handle deserialization
        # difference (`_id` vs `id`)
        field_kwargs = self._extract_marshmallow_field_params(mongo_world)
        if self.value_field:
            inner_ma_schema = self.value_field.as_marshmallow_field(
                mongo_world=mongo_world, **kwargs)
        else:
            inner_ma_schema = None
        return ma_fields.Dict(self.key_field, inner_ma_schema, **field_kwargs)

    def _required_validate(self, value):
        if value is missing or not hasattr(self.value_field, '_required_validate'):
            return
        required_validate = self.value_field._required_validate
        errors = collections.defaultdict(dict)
        for key, val in value.items():
            try:
                required_validate(val)
            except ValidationError as exc:
                errors[key]["value"] = exc.messages
        if errors:
            raise ValidationError(errors)


class ListField(BaseField, ma_fields.List):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        def cast_value_or_callable(inner, value):
            if value is missing:
                return missing
            if callable(value):
                return lambda: List(inner, value())
            return List(inner, value)

        self.default = cast_value_or_callable(self.inner, self.default)
        self.missing = cast_value_or_callable(self.inner, self.missing)

    def _deserialize(self, value, attr, data, **kwargs):
        value = super()._deserialize(value, attr, data, **kwargs)
        return List(self.inner, value)

    def _serialize_to_mongo(self, obj):
        if obj is None:
            return missing
        return [self.inner.serialize_to_mongo(each) for each in obj]

    def _deserialize_from_mongo(self, value):
        if value:
            return List(
                self.inner,
                [self.inner.deserialize_from_mongo(each) for each in value]
            )
        return List(self.inner)

    def map_to_field(self, mongo_path, path, func):
        """Apply a function to every field in the schema
        """
        func(mongo_path, path, self.inner)
        if hasattr(self.inner, 'map_to_field'):
            self.inner.map_to_field(mongo_path, path, func)

    def as_marshmallow_field(self, mongo_world=False, **kwargs):
        # Overwrite default `as_marshmallow_field` to handle deserialization
        # difference (`_id` vs `id`)
        field_kwargs = self._extract_marshmallow_field_params(mongo_world)
        inner_ma_schema = self.inner.as_marshmallow_field(
            mongo_world=mongo_world, **kwargs)
        return ma_fields.List(inner_ma_schema, **field_kwargs)

    def _required_validate(self, value):
        if value is missing or not hasattr(self.inner, '_required_validate'):
            return
        required_validate = self.inner._required_validate
        errors = {}
        for i, sub_value in enumerate(value):
            try:
                required_validate(sub_value)
            except ValidationError as exc:
                errors[i] = exc.messages
        if errors:
            raise ValidationError(errors)


# Aliases
URLField = UrlField
StrField = StringField
BoolField = BooleanField
IntField = IntegerField


class ObjectIdField(BaseField, ma_bonus_fields.ObjectId):
    pass


class ReferenceField(BaseField, ma_bonus_fields.Reference):

    def __init__(self, document, *args, reference_cls=Reference, **kwargs):
        """
        :param document: Can be a :class:`umongo.embedded_document.DocumentTemplate`,
            another instance's :class:`umongo.embedded_document.DocumentImplementation` or
            the embedded document class name.
        """
        super().__init__(*args, **kwargs)
        # TODO : check document_cls is implementation or string
        self.reference_cls = reference_cls
        # Can be the Template, Template's name or another Implementation
        if not isinstance(document, str):
            self.document = get_template(document)
        else:
            self.document = document
        self._document_cls = None
        # Avoid importing multiple times
        from .document import DocumentImplementation
        self._document_implementation_cls = DocumentImplementation

    @property
    def document_cls(self):
        """
        Return the instance's :class:`umongo.embedded_document.DocumentImplementation`
        implementing the `document` attribute.
        """
        if not self._document_cls:
            self._document_cls = self.instance.retrieve_document(self.document)
        return self._document_cls

    def _deserialize(self, value, attr, data, **kwargs):
        if value is None:
            return None
        if isinstance(value, DBRef):
            if self._document_cls.collection.name != value.collection:
                raise ValidationError(_("DBRef must be on collection `{collection}`.").format(
                    self._document_cls.collection.name))
            value = value.id
        elif isinstance(value, Reference):
            if value.document_cls != self.document_cls:
                raise ValidationError(_("`{document}` reference expected.").format(
                    document=self.document_cls.__name__))
            if not isinstance(value, self.reference_cls):
                value = self.reference_cls(value.document_cls, value.pk)
            return value
        elif isinstance(value, self.document_cls):
            if not value.is_created:
                raise ValidationError(
                    _("Cannot reference a document that has not been created yet."))
            value = value.pk
        elif isinstance(value, self._document_implementation_cls):
            raise ValidationError(_("`{document}` reference expected.").format(
                document=self.document_cls.__name__))
        value = super()._deserialize(value, attr, data, **kwargs)
        return self.reference_cls(self.document_cls, value)

    def _serialize_to_mongo(self, obj):
        return obj.pk

    def _deserialize_from_mongo(self, value):
        return self.reference_cls(self.document_cls, value)

    def as_marshmallow_field(self, mongo_world=False, **kwargs):
        # Overwrite default `as_marshmallow_field` to handle deserialization
        # difference (`_id` vs `id`)
        field_kwargs = self._extract_marshmallow_field_params(mongo_world)
        return ma_bonus_fields.Reference(mongo_world=mongo_world, **field_kwargs)


class GenericReferenceField(BaseField, ma_bonus_fields.GenericReference):

    def __init__(self, *args, reference_cls=Reference, **kwargs):
        super().__init__(*args, **kwargs)
        self.reference_cls = reference_cls
        self._document_implementation_cls = DocumentImplementation

    def _document_cls(self, class_name):
        try:
            return self.instance.retrieve_document(class_name)
        except NotRegisteredDocumentError:
            raise ValidationError(_('Unknown document `{document}`.').format(
                document=class_name))

    def _serialize(self, value, attr, obj):
        if value is None:
            return None
        return {'id': str(value.pk), 'cls': value.document_cls.__name__}

    def _deserialize(self, value, attr, data, **kwargs):
        if value is None:
            return None
        if isinstance(value, Reference):
            if not isinstance(value, self.reference_cls):
                value = self.reference_cls(value.document_cls, value.pk)
            return value
        if isinstance(value, self._document_implementation_cls):
            if not value.is_created:
                raise ValidationError(
                    _("Cannot reference a document that has not been created yet."))
            return self.reference_cls(value.__class__, value.pk)
        if isinstance(value, dict):
            if value.keys() != {'cls', 'id'}:
                raise ValidationError(_("Generic reference must have `id` and `cls` fields."))
            try:
                _id = ObjectId(value['id'])
            except ValueError:
                raise ValidationError(_("Invalid `id` field."))
            document_cls = self._document_cls(value['cls'])
            return self.reference_cls(document_cls, _id)
        raise ValidationError(_("Invalid value for generic reference field."))

    def _serialize_to_mongo(self, obj):
        return {'_id': obj.pk, '_cls': obj.document_cls.__name__}

    def _deserialize_from_mongo(self, value):
        document_cls = self._document_cls(value['_cls'])
        return self.reference_cls(document_cls, value['_id'])

    def as_marshmallow_field(self, mongo_world=False, **kwargs):
        # Overwrite default `as_marshmallow_field` to handle deserialization
        # difference (`_id` vs `id`)
        field_kwargs = self._extract_marshmallow_field_params(mongo_world)
        return ma_bonus_fields.GenericReference(mongo_world=mongo_world, **field_kwargs)


class EmbeddedField(BaseField, ma_fields.Nested):

    def __init__(self, embedded_document, *args, **kwargs):
        """
        :param embedded_document: Can be a
            :class:`umongo.embedded_document.EmbeddedDocumentTemplate`,
            another instance's :class:`umongo.embedded_document.EmbeddedDocumentImplementation`
            or the embedded document class name.
        """
        # Don't need to pass `nested` attribute given it is overloaded
        super().__init__(None, *args, **kwargs)
        # Try to retrieve the template if possible for consistency
        if not isinstance(embedded_document, str):
            self.embedded_document = get_template(embedded_document)
        else:
            self.embedded_document = embedded_document
        self._embedded_document_cls = None

    @property
    def nested(self):
        # Overload `nested` attribute to be able to fetch it lazily
        return self.embedded_document_cls.Schema

    @nested.setter
    def nested(self, value):
        pass

    @property
    def embedded_document_cls(self):
        """
        Return the instance's :class:`umongo.embedded_document.EmbeddedDocumentImplementation`
        implementing the `embedded_document` attribute.
        """
        if not self._embedded_document_cls:
            self._embedded_document_cls = self.instance.retrieve_embedded_document(
                self.embedded_document)
        return self._embedded_document_cls

    def _serialize(self, value, attr, obj):
        if value is None:
            return None
        return value.dump()

    def _deserialize(self, value, attr, data, **kwargs):
        embedded_document_cls = self.embedded_document_cls
        if isinstance(value, embedded_document_cls):
            return value
        if not isinstance(value, dict):
            raise ValidationError({'_schema': ['Invalid input type.']})
        # Handle inheritance deserialization here using `cls` field as hint
        if embedded_document_cls.opts.offspring and 'cls' in value:
            to_use_cls_name = value.pop('cls')
            if not any(o for o in embedded_document_cls.opts.offspring
                       if o.__name__ == to_use_cls_name):
                raise ValidationError(_('Unknown document `{document}`.').format(
                    document=to_use_cls_name))
            try:
                to_use_cls = embedded_document_cls.opts.instance.retrieve_embedded_document(
                    to_use_cls_name)
            except NotRegisteredDocumentError as exc:
                raise ValidationError(str(exc))
            return to_use_cls(**value)
        return embedded_document_cls(**value)

    def _serialize_to_mongo(self, obj):
        return obj.to_mongo()

    def _deserialize_from_mongo(self, value):
        return self.embedded_document_cls.build_from_mongo(value)

    def _validate_missing(self, value):
        # Overload default to handle recursive check
        super()._validate_missing(value)
        errors = {}
        if value is missing:
            def get_sub_value(_):
                return missing
        elif isinstance(value, dict):
            # value is a dict for deserialization
            def get_sub_value(key):
                return value.get(key, missing)
        elif isinstance(value, self.embedded_document_cls):
            # value is a valid EmbeddedDocument
            def get_sub_value(key):
                return value._data.get(key)
        else:
            # value is invalid, just return and let `_deserialize`
            # raises an error about this
            return
        for name, field in self.embedded_document_cls.schema.fields.items():
            sub_value = get_sub_value(name)
            # `_validate_missing` doesn't check for required fields here, so we
            # can safely skip missing values
            if sub_value is missing:
                continue
            try:
                field._validate_missing(sub_value)
            except ValidationError as exc:
                errors[name] = exc.messages
        if errors:
            raise ValidationError(errors)

    def map_to_field(self, mongo_path, path, func):
        """Apply a function to every field in the schema"""
        for name, field in self.embedded_document_cls.schema.fields.items():
            cur_path = '%s.%s' % (path, name)
            cur_mongo_path = '%s.%s' % (mongo_path, field.attribute or name)
            func(cur_mongo_path, cur_path, field)
            if hasattr(field, 'map_to_field'):
                field.map_to_field(cur_mongo_path, cur_path, func)

    def as_marshmallow_field(self, mongo_world=False, **kwargs):
        # Overwrite default `as_marshmallow_field` to handle nesting
        field_kwargs = self._extract_marshmallow_field_params(mongo_world)
        nested_ma_schema = self._embedded_document_cls.schema.as_marshmallow_schema(
            mongo_world=mongo_world)
        return ma_fields.Nested(nested_ma_schema, **field_kwargs)

    def _required_validate(self, value):
        if value is not missing:
            value.required_validate()



import pytest

from umongo import Document, fields, exceptions

from .common import BaseTest


class TestInheritance(BaseTest):

    def test_cls_field(self):

        @self.instance.register
        class Parent(Document):
            last_name = fields.StrField()

            class Meta:
                allow_inheritance = True

        @self.instance.register
        class Child(Parent):
            first_name = fields.StrField()

        assert 'cls' in Child.schema.fields
        Child.schema.fields['cls']
        assert not hasattr(Parent(), 'cls')
        assert Child().cls == 'Child'

        loaded = Parent.build_from_mongo(
            {'_cls': 'Child', 'first_name': 'John', 'last_name': 'Doe'}, use_cls=True)
        assert loaded.cls == 'Child'

    def test_simple(self):

        @self.instance.register
        class Parent(Document):
            last_name = fields.StrField()

            class Meta:
                allow_inheritance = True
                collection_name = 'parent_col'

        assert Parent.opts.abstract is False
        assert Parent.opts.allow_inheritance is True

        @self.instance.register
        class Child(Parent):
            first_name = fields.StrField()

        assert Child.opts.abstract is False
        assert Child.opts.allow_inheritance is False
        assert Child.opts.collection_name == 'parent_col'
        assert Child.collection.name == 'parent_col'
        Child(first_name='John', last_name='Doe')

    def test_abstract(self):

        # Cannot define a collection_name for an abstract doc !
        with pytest.raises(exceptions.DocumentDefinitionError):
            @self.instance.register
            class BadAbstractDoc(Document):
                class Meta:
                    abstract = True
                    collection_name = 'my_col'

        @self.instance.register
        class AbstractDoc(Document):
            abs_field = fields.StrField(default='from abstract')

            class Meta:
                abstract = True

        assert AbstractDoc.opts.abstract is True
        assert AbstractDoc.opts.allow_inheritance is True
        # Cannot instanciate also an abstract document
        with pytest.raises(exceptions.AbstractDocumentError):
            AbstractDoc()

        @self.instance.register
        class StillAbstractDoc(AbstractDoc):
            class Meta:
                abstract = True

        assert StillAbstractDoc.opts.abstract is True
        assert StillAbstractDoc.opts.allow_inheritance is True

        @self.instance.register
        class ConcreteDoc(AbstractDoc):
            pass

        assert ConcreteDoc.opts.abstract is False
        assert ConcreteDoc.opts.allow_inheritance is False
        assert ConcreteDoc().abs_field == 'from abstract'

    def test_non_document_inheritance(self):

        class NotDoc1:
            @staticmethod
            def my_func1():
                return 24

        class NotDoc2:
            @staticmethod
            def my_func2():
                return 42

        @self.instance.register
        class Doc(NotDoc1, Document, NotDoc2):
            a = fields.StrField()

        assert issubclass(Doc, NotDoc1)
        assert issubclass(Doc, NotDoc2)
        assert isinstance(Doc(), NotDoc1)
        assert isinstance(Doc(), NotDoc2)
        assert Doc.my_func1() == 24
        assert Doc.my_func2() == 42
        doc = Doc(a='test')
        assert doc.my_func1() == 24
        assert doc.my_func2() == 42
        assert doc.a == 'test'
